from io import BytesIO
from random import randint
from uuid import uuid4

import grpc
import pytest
from mock import AsyncMock, MagicMock, patch

from pyzeebe.errors import (InvalidJSONError,
                            ProcessDefinitionHasNoStartEventError,
                            ProcessDefinitionNotFoundError,
                            ProcessInstanceNotFoundError, ProcessInvalidError)
from pyzeebe.grpc_internals.zeebe_process_adapter import ZeebeProcessAdapter
from tests.unit.utils.gateway_mock import GatewayMock
from tests.unit.utils.grpc_utils import GRPCStatusCode
from tests.unit.utils.random_utils import RANDOM_RANGE


@pytest.mark.asyncio
class TestCreateProcessInstance:
    async def test_response_is_of_correct_type(self, zeebe_adapter: ZeebeProcessAdapter, grpc_servicer: GatewayMock):
        bpmn_process_id = str(uuid4())
        version = randint(0, 10)
        grpc_servicer.mock_deploy_process(bpmn_process_id, version, [])

        response = await zeebe_adapter.create_process_instance(
            bpmn_process_id, version, {}
        )

        assert isinstance(response, int)

    async def test_raises_on_unkown_process(self, zeebe_adapter: ZeebeProcessAdapter):
        bpmn_process_id = str(uuid4())
        version = randint(0, 10)

        with pytest.raises(ProcessDefinitionNotFoundError):
            await zeebe_adapter.create_process_instance(bpmn_process_id, version, {})

    async def test_raises_on_invalid_json(self, zeebe_adapter: ZeebeProcessAdapter):
        bpmn_process_id = str(uuid4())
        version = randint(0, 10)

        error = grpc.aio.AioRpcError(
            grpc.StatusCode.INVALID_ARGUMENT, None, None
        )
        zeebe_adapter._gateway_stub.CreateProcessInstance = AsyncMock(
            side_effect=error
        )

        with pytest.raises(InvalidJSONError):
            await zeebe_adapter.create_process_instance(bpmn_process_id, version, None)

    async def test_raises_on_no_start_event(self, zeebe_adapter: ZeebeProcessAdapter):
        bpmn_process_id = str(uuid4())
        version = randint(0, 10)
        error = grpc.aio.AioRpcError(
            grpc.StatusCode.FAILED_PRECONDITION, None, None
        )
        zeebe_adapter._gateway_stub.CreateProcessInstance = AsyncMock(
            side_effect=error
        )

        with pytest.raises(ProcessDefinitionHasNoStartEventError):
            await zeebe_adapter.create_process_instance(bpmn_process_id, version, {})


@pytest.mark.asyncio
class TestCreateProcessWithResult:
    async def test_process_instance_key_type_is_int(self, zeebe_adapter: ZeebeProcessAdapter, grpc_servicer: GatewayMock):
        bpmn_process_id = str(uuid4())
        version = randint(0, 10)
        grpc_servicer.mock_deploy_process(bpmn_process_id, version, [])

        process_instance_key, _ = await zeebe_adapter.create_process_instance_with_result(
            bpmn_process_id=bpmn_process_id,
            variables={},
            version=version,
            timeout=0,
            variables_to_fetch=[]
        )

        assert isinstance(process_instance_key, int)

    async def test_variables_type_is_dict(self, zeebe_adapter: ZeebeProcessAdapter, grpc_servicer: GatewayMock):
        bpmn_process_id = str(uuid4())
        version = randint(0, 10)
        grpc_servicer.mock_deploy_process(bpmn_process_id, version, [])

        _, response = await zeebe_adapter.create_process_instance_with_result(
            bpmn_process_id=bpmn_process_id,
            variables={},
            version=version,
            timeout=0,
            variables_to_fetch=[]
        )

        assert isinstance(response, dict)


@pytest.mark.asyncio
class TestCancelProcess:
    async def test_cancels_the_process(self, zeebe_adapter: ZeebeProcessAdapter, grpc_servicer: GatewayMock):
        bpmn_process_id = str(uuid4())
        version = randint(0, 10)
        grpc_servicer.mock_deploy_process(bpmn_process_id, version, [])
        process_instance_key = await zeebe_adapter.create_process_instance(
            bpmn_process_id=bpmn_process_id,
            variables={},
            version=version
        )

        await zeebe_adapter.cancel_process_instance(process_instance_key)

        assert process_instance_key not in grpc_servicer.active_processes.keys()

    async def test_raises_on_already_cancelled_process(self, zeebe_adapter: ZeebeProcessAdapter, grpc_servicer: GatewayMock):
        error = grpc.aio.AioRpcError(
            grpc.StatusCode.NOT_FOUND, None, None
        )

        zeebe_adapter._gateway_stub.CancelProcessInstance = AsyncMock(
            side_effect=error
        )

        with pytest.raises(ProcessInstanceNotFoundError):
            await zeebe_adapter.cancel_process_instance(
                process_instance_key=randint(0, RANDOM_RANGE)
            )


@pytest.mark.asyncio
class TestDeployProcess:
    open_mock: MagicMock

    @pytest.fixture(autouse=True)
    def mocked_aiofiles_open(self):
        read_mock = AsyncMock(return_value=bytes())

        file_mock = AsyncMock()
        file_mock.__aenter__.return_value.read = read_mock

        with patch("pyzeebe.grpc_internals.zeebe_process_adapter.aiofiles.open", return_value=file_mock) as open_mock:
            self.open_mock = open_mock
            yield

    async def test_raises_on_invalid_process(self, zeebe_adapter: ZeebeProcessAdapter):
        error = grpc.aio.AioRpcError(
            grpc.StatusCode.INVALID_ARGUMENT, None, None
        )

        zeebe_adapter._gateway_stub.DeployProcess = AsyncMock(
            side_effect=error
        )

        with pytest.raises(ProcessInvalidError):
            await zeebe_adapter.deploy_process()

    async def test_calls_open_in_rb_mode(self, zeebe_adapter: ZeebeProcessAdapter):
        file_path = str(uuid4())

        await zeebe_adapter.deploy_process(file_path)

        self.open_mock.assert_called_with(file_path, "rb")



from uuid import uuid4

import grpc
import pytest
from mock import AsyncMock

from pyzeebe import SyncZeebeClient
from pyzeebe.errors import ProcessDefinitionNotFoundError
from pyzeebe.grpc_internals.zeebe_adapter import ZeebeAdapter

# Pytest doesn't play well with loop.run_until_complete unless the test has a
# running asyncio loop
pytestmark = pytest.mark.asyncio


@pytest.fixture
def sync_zeebe_client(aio_grpc_channel: grpc.aio.Channel) -> SyncZeebeClient:
    client = SyncZeebeClient(aio_grpc_channel)
    return client


class TestRunProcess:
    def test_run_process_returns_int(self, sync_zeebe_client: SyncZeebeClient, deployed_process):
        bpmn_process_id, version = deployed_process

        process_instance_key = sync_zeebe_client.run_process(
            bpmn_process_id, version=version
        )

        assert isinstance(process_instance_key, int)

    def test_raises_process_definition_not_found_error_for_invalid_process_id(self, sync_zeebe_client: SyncZeebeClient):
        with pytest.raises(ProcessDefinitionNotFoundError):
            sync_zeebe_client.run_process(bpmn_process_id=str(uuid4()))


class TestRunProcessWithResult:
    def test_run_process_with_result_instance_key_is_int(self, sync_zeebe_client: SyncZeebeClient, deployed_process):
        bpmn_process_id, version = deployed_process

        process_instance_key, _ = sync_zeebe_client.run_process_with_result(
            bpmn_process_id, {}, version
        )

        assert isinstance(process_instance_key, int)

    def test_run_process_with_result_output_variables_are_as_expected(self, sync_zeebe_client: SyncZeebeClient, deployed_process):
        expected = {}
        bpmn_process_id, version = deployed_process

        _, output_variables = sync_zeebe_client.run_process_with_result(
            bpmn_process_id, {}, version
        )

        assert output_variables == expected

    def test_raises_process_definition_not_found_error_for_invalid_process_id(self, sync_zeebe_client: SyncZeebeClient):
        with pytest.raises(ProcessDefinitionNotFoundError):
            sync_zeebe_client.run_process_with_result(
                bpmn_process_id=str(uuid4())
            )


class TestCancelProcessInstance:
    def test_cancel_process_instance(self, sync_zeebe_client: SyncZeebeClient, deployed_process):
        bpmn_process_id, version = deployed_process
        process_instance_key = sync_zeebe_client.run_process(
            bpmn_process_id=bpmn_process_id, variables={}, version=version
        )

        returned_process_instance_key = sync_zeebe_client.cancel_process_instance(
            process_instance_key
        )

        assert returned_process_instance_key == process_instance_key


class TestDeployProcess:
    def test_calls_deploy_process_of_zeebe_adapter(self, sync_zeebe_client: SyncZeebeClient):
        sync_zeebe_client.zeebe_adapter.deploy_process = AsyncMock()
        file_path = str(uuid4())

        sync_zeebe_client.deploy_process(file_path)

        sync_zeebe_client.zeebe_adapter.deploy_process.assert_called_with(
            file_path
        )


class TestPublishMessage:
    def test_calls_publish_message_of_zeebe_adapter(self, sync_zeebe_client: SyncZeebeClient):
        sync_zeebe_client.zeebe_adapter.publish_message = AsyncMock()
        name = str(uuid4())
        correlation_key = str(uuid4())

        sync_zeebe_client.publish_message(name, correlation_key)

        sync_zeebe_client.zeebe_adapter.publish_message.assert_called_once()



from typing import Dict, Optional

import grpc
from oauthlib import oauth2
from requests import HTTPError
from requests_oauthlib import OAuth2Session

from pyzeebe.channel.channel_options import get_channel_options
from pyzeebe.errors import (InvalidCamundaCloudCredentialsError,
                            InvalidOAuthCredentialsError)


def create_camunda_cloud_channel(
    client_id: str,
    client_secret: str,
    cluster_id: str,
    channel_options: Optional[Dict] = None,
) -> grpc.aio.Channel:
    """
    Create channel connected to a Camunda Cloud cluster

    Args:
        client_id (str): The client id provided by Camunda Cloud
        client_secret (str): The client secret provided by Camunda Cloud
        cluster_id (str): The zeebe cluster id to connect to
        channel_options (Optional[Dict], optional): GRPC channel options. See https://grpc.github.io/grpc/python/glossary.html

    Returns:
        grpc.aio.Channel: A GRPC Channel connected to the Zeebe gateway.

    Raises:
        InvalidCamundaCloudCredentialsError: One of the provided camunda credentials is not correct
    """
    channel_credentials = _create_camunda_cloud_credentials(
        client_id, client_secret, cluster_id
    )

    return grpc.aio.secure_channel(
        f"{cluster_id}.zeebe.camunda.io:443",
        channel_credentials,
        options=get_channel_options(channel_options),
    )


def _create_camunda_cloud_credentials(
    client_id: str, client_secret: str, cluster_id: str
) -> grpc.ChannelCredentials:
    try:
        access_token = _get_access_token(
            "https://login.cloud.camunda.io/oauth/token",
            client_id,
            client_secret,
            cluster_id,
        )
        return _create_oauth_credentials(access_token)
    except InvalidOAuthCredentialsError as oauth_error:
        raise InvalidCamundaCloudCredentialsError(
            client_id, cluster_id
        ) from oauth_error


def _get_access_token(
    url: str, client_id: str, client_secret: str, audience: str
) -> str:
    try:
        client = oauth2.BackendApplicationClient(client_id)
        client.prepare_request_body(include_client_id=True)
        with OAuth2Session(client=client) as session:
            response = session.post(
                url,
                data={
                    "client_id": client_id,
                    "client_secret": client_secret,
                    "audience": audience,
                },
            )
            response.raise_for_status()
            return response.json()["access_token"]
    except HTTPError as http_error:
        raise InvalidOAuthCredentialsError(
            url=url, client_id=client_id, audience=audience
        ) from http_error


def _create_oauth_credentials(access_token: str) -> grpc.ChannelCredentials:
    token_credentials = grpc.access_token_call_credentials(access_token)
    ssl_credentials = grpc.ssl_channel_credentials()
    return grpc.composite_channel_credentials(ssl_credentials, token_credentials)



import asyncio

import pytest
from mock import AsyncMock, Mock

from pyzeebe.job.job import Job
from pyzeebe.task.task import Task
from pyzeebe.worker.job_executor import JobExecutor, create_job_callback
from pyzeebe.worker.task_state import TaskState


@pytest.fixture
def job_executor(task: Task, queue: asyncio.Queue, task_state: TaskState):
    return JobExecutor(task, queue, task_state)


@pytest.fixture(autouse=True)
def mock_job_handler(task: Task):
    task.job_handler = AsyncMock()


@pytest.mark.asyncio
class TestExecuteOneJob:
    async def test_executes_jobs(self, job_executor: JobExecutor, job_from_task: Job, task: Task):
        await job_executor.execute_one_job(job_from_task)

        task.job_handler.assert_called_with(job_from_task)

    async def test_continues_on_deactivated_job(self, job_executor: JobExecutor, job_from_task: Job):
        await job_executor.execute_one_job(job_from_task)
        await job_executor.execute_one_job(job_from_task)


@pytest.mark.asyncio
class TestGetNextJob:
    async def test_returns_expected_job(self, job_executor: JobExecutor, job_from_task: Job):
        await job_executor.jobs.put(job_from_task)

        assert await job_executor.get_next_job() == job_from_task


class TestShouldExecute:
    def test_returns_true_on_default(self, job_executor: JobExecutor):
        assert job_executor.should_execute()

    @pytest.mark.asyncio
    async def test_returns_false_when_executor_is_stopped(self, job_executor: JobExecutor):
        await job_executor.stop()

        assert not job_executor.should_execute()


@pytest.mark.asyncio
class TestStop:
    async def test_stops_executor(self, job_executor: JobExecutor):
        await job_executor.stop()

        await job_executor.execute()  # Implicitly test that execute returns immediately


class TestCreateJobCallback:
    def test_returns_callable(self, job_executor: JobExecutor, job_from_task: Job):
        callback = create_job_callback(job_executor, job_from_task)

        assert callable(callback)

    def test_signals_that_job_is_done(self, job_executor: JobExecutor, job_from_task: Job):
        task_done_mock = Mock()
        remove_from_task_state_mock = Mock()
        job_executor.jobs.task_done = task_done_mock
        job_executor.task_state.remove = remove_from_task_state_mock

        callback = create_job_callback(job_executor, job_from_task)
        callback(asyncio.Future())

        task_done_mock.assert_called_once()
        remove_from_task_state_mock.assert_called_once_with(job_from_task)



import asyncio
import os
from typing import Dict
from uuid import uuid4

import grpc
import pytest

from pyzeebe import Job, ZeebeClient, ZeebeWorker, create_insecure_channel
from pyzeebe.errors import ProcessDefinitionNotFoundError


@pytest.fixture(scope="module")
def event_loop():
    loop = asyncio.new_event_loop()
    yield loop
    loop.close()


@pytest.fixture(scope="module")
def grpc_channel():
    return create_insecure_channel()


@pytest.fixture(scope="module")
def zeebe_client(grpc_channel: grpc.aio.Channel):
    return ZeebeClient(grpc_channel)


@pytest.fixture(scope="module")
def zeebe_worker(grpc_channel):
    return ZeebeWorker(grpc_channel)


@pytest.fixture(autouse=True, scope="module")
def task(zeebe_worker: ZeebeWorker):
    def exception_handler(exc: Exception, job: Job) -> None:
        job.set_error_status(f"Failed to run task {job.type}. Reason: {exc}")

    @zeebe_worker.task("test", exception_handler)
    async def task_handler(should_throw: bool, input: str) -> Dict:
        if should_throw:
            raise Exception("Error thrown")
        else:
            return {"output": input + str(uuid4())}


@pytest.mark.asyncio
@pytest.fixture(autouse=True, scope="module")
async def deploy_process(zeebe_client: ZeebeClient):
    try:
        integration_tests_path = os.path.join("tests", "integration")
        await zeebe_client.deploy_process(
            os.path.join(integration_tests_path, "test.bpmn")
        )
    except FileNotFoundError:
        await zeebe_client.deploy_process("test.bpmn")


@pytest.fixture(autouse=True, scope="module")
def start_worker(event_loop: asyncio.AbstractEventLoop, zeebe_worker: ZeebeWorker):
    event_loop.create_task(zeebe_worker.work())
    yield
    event_loop.create_task(zeebe_worker.stop())


@pytest.mark.asyncio
async def test_run_process(zeebe_client: ZeebeClient):
    process_key = await zeebe_client.run_process(
        "test", {"input": str(uuid4()), "should_throw": False}
    )
    assert isinstance(process_key, int)


@pytest.mark.asyncio
async def test_non_existent_process(zeebe_client: ZeebeClient):
    with pytest.raises(ProcessDefinitionNotFoundError):
        await zeebe_client.run_process(str(uuid4()))


@pytest.mark.asyncio
async def test_run_process_with_result(zeebe_client: ZeebeClient):
    input = str(uuid4())
    process_instance_key, process_result = await zeebe_client.run_process_with_result(
        "test", {"input": input, "should_throw": False}
    )
    assert isinstance(process_instance_key, int)
    assert isinstance(process_result["output"], str)
    assert process_result["output"].startswith(input)


@pytest.mark.asyncio
async def test_cancel_process(zeebe_client: ZeebeClient):
    process_key = await zeebe_client.run_process(
        "test", {"input": str(uuid4()), "should_throw": False}
    )
    await zeebe_client.cancel_process_instance(process_key)



from random import randint
from uuid import uuid4

import pytest
from mock import AsyncMock

from pyzeebe.errors import ProcessDefinitionNotFoundError


@pytest.mark.asyncio
async def test_run_process(zeebe_client, grpc_servicer):
    bpmn_process_id = str(uuid4())
    version = randint(0, 10)
    grpc_servicer.mock_deploy_process(bpmn_process_id, version, [])
    assert isinstance(await zeebe_client.run_process(
        bpmn_process_id=bpmn_process_id, variables={}, version=version), int)


@pytest.mark.asyncio
class TestRunProcessWithResult:
    async def test_run_process_with_result_instance_key_is_int(self, zeebe_client, deployed_process):
        bpmn_process_id, version = deployed_process

        process_instance_key, _ = await zeebe_client.run_process_with_result(
            bpmn_process_id, {}, version
        )

        assert isinstance(process_instance_key, int)

    async def test_run_process_with_result_output_variables_are_as_expected(self, zeebe_client, deployed_process):
        expected = {}
        bpmn_process_id, version = deployed_process

        _, output_variables = await zeebe_client.run_process_with_result(
            bpmn_process_id, {}, version
        )

        assert output_variables == expected


@pytest.mark.asyncio
async def test_deploy_process(zeebe_client):
    zeebe_client.zeebe_adapter.deploy_process = AsyncMock()
    file_path = str(uuid4())
    await zeebe_client.deploy_process(file_path)
    zeebe_client.zeebe_adapter.deploy_process.assert_called_with(file_path)


@pytest.mark.asyncio
async def test_run_non_existent_process(zeebe_client):
    with pytest.raises(ProcessDefinitionNotFoundError):
        await zeebe_client.run_process(bpmn_process_id=str(uuid4()))


@pytest.mark.asyncio
async def test_run_non_existent_process_with_result(zeebe_client):
    with pytest.raises(ProcessDefinitionNotFoundError):
        await zeebe_client.run_process_with_result(bpmn_process_id=str(uuid4()))


@pytest.mark.asyncio
async def test_cancel_process_instance(zeebe_client, grpc_servicer):
    bpmn_process_id = str(uuid4())
    version = randint(0, 10)
    grpc_servicer.mock_deploy_process(bpmn_process_id, version, [])
    process_instance_key = await zeebe_client.run_process(
        bpmn_process_id=bpmn_process_id, variables={}, version=version)
    assert isinstance(await zeebe_client.cancel_process_instance(process_instance_key=process_instance_key), int)


@pytest.mark.asyncio
async def test_publish_message(zeebe_client):
    await zeebe_client.publish_message(name=str(uuid4()), correlation_key=str(uuid4()))



from typing import List, Optional

from pyzeebe.errors import NoVariableNameGivenError
from pyzeebe.function_tools import async_tools
from pyzeebe.task.exception_handler import ExceptionHandler
from pyzeebe.task.types import AsyncTaskDecorator, TaskDecorator


class TaskConfig:
    before: List[AsyncTaskDecorator]
    after: List[AsyncTaskDecorator]

    def __init__(self, type: str, exception_handler: ExceptionHandler,
                 timeout_ms: int, max_jobs_to_activate: int,
                 variables_to_fetch: List[str],
                 single_value: bool, variable_name: str, before: List[TaskDecorator],
                 after: List[TaskDecorator]):
        if single_value and not variable_name:
            raise NoVariableNameGivenError(type)

        self.type = type
        self.exception_handler = exception_handler
        self.timeout_ms = timeout_ms
        self.max_jobs_to_activate = max_jobs_to_activate
        self.variables_to_fetch = variables_to_fetch
        self.single_value = single_value
        self.variable_name = variable_name
        self.before = async_tools.asyncify_all_functions(before)
        self.after = async_tools.asyncify_all_functions(after)
        self.job_parameter_name: Optional[str] = None

    def __repr__(self):
        return f"TaskConfig(type={self.type}, exception_handler={self.exception_handler}, " \
               f"timeout_ms={self.timeout_ms}, max_jobs_to_activate={self.max_jobs_to_activate}, " \
               f"variables_to_fetch={self.variables_to_fetch}, single_value={self.single_value}, " \
               f"variable_name={self.variable_name}, before={self.before}, after={self.after})"



class PyZeebeError(Exception):
    pass


class TaskNotFoundError(PyZeebeError):
    pass


class NoVariableNameGivenError(PyZeebeError):
    def __init__(self, task_type: str):
        super().__init__(f"No variable name given for single_value task {task_type}")
        self.task_type = task_type


class NoZeebeAdapterError(PyZeebeError):
    pass


class DuplicateTaskTypeError(PyZeebeError):
    def __init__(self, task_type: str):
        super().__init__(f"Task with type {task_type} already exists")
        self.task_type = task_type


class MaxConsecutiveTaskThreadError(PyZeebeError):
    pass


class BusinessError(PyZeebeError):
    """
    Exception that can be raised with a user defined code,
    to be caught later by an error event in the Zeebe process
    """

    def __init__(self, error_code: str, msg: str = None) -> None:
        if msg is None:
            msg = f"Business error with code {error_code}"
        super().__init__(msg)
        self.error_code = error_code



from pyzeebe.errors.pyzeebe_errors import PyZeebeError


class ZeebeBackPressureError(PyZeebeError):
    pass


class ZeebeGatewayUnavailableError(PyZeebeError):
    pass


class ZeebeInternalError(PyZeebeError):
    pass



import asyncio
import logging
from typing import Callable

from pyzeebe.errors import JobAlreadyDeactivatedError
from pyzeebe.job.job import Job
from pyzeebe.task.task import Task
from pyzeebe.worker.task_state import TaskState

logger = logging.getLogger(__name__)

AsyncTaskCallback = Callable[[asyncio.Future], None]


class JobExecutor:
    def __init__(self, task: Task, jobs: asyncio.Queue, task_state: TaskState):
        self.task = task
        self.jobs = jobs
        self.task_state = task_state
        self.stop_event = asyncio.Event()

    async def execute(self) -> None:
        while self.should_execute():
            job = await self.get_next_job()
            task = asyncio.create_task(
                self.execute_one_job(job)
            )
            task.add_done_callback(create_job_callback(self, job))

    async def get_next_job(self) -> Job:
        return await self.jobs.get()

    async def execute_one_job(self, job: Job) -> None:
        try:
            await self.task.job_handler(job)
        except JobAlreadyDeactivatedError as error:
            logger.warning(
                f"Job was already deactivated. Job key: {error.job_key}"
            )

    def should_execute(self) -> bool:
        return not self.stop_event.is_set()

    async def stop(self) -> None:
        self.stop_event.set()
        await self.jobs.join()


def create_job_callback(job_executor: JobExecutor, job: Job) -> AsyncTaskCallback:
    def callback(future: asyncio.Future):
        job_executor.jobs.task_done()
        job_executor.task_state.remove(job)

    return callback



import asyncio

import pytest

from pyzeebe.grpc_internals.zeebe_adapter import ZeebeAdapter
from pyzeebe.job.job import Job
from pyzeebe.task.task import Task
from pyzeebe.worker.job_poller import JobPoller
from pyzeebe.worker.task_state import TaskState
from tests.unit.utils.gateway_mock import GatewayMock
from tests.unit.utils.random_utils import random_job


@pytest.fixture
def job_poller(zeebe_adapter: ZeebeAdapter, task: Task, queue: asyncio.Queue, task_state: TaskState) -> JobPoller:
    return JobPoller(zeebe_adapter, task, queue, "test_worker", 100, task_state, 10)


@pytest.mark.asyncio
class TestPollOnce:
    async def test_one_job_is_polled(self, job_poller: JobPoller, queue: asyncio.Queue, job_from_task: Job,
                                     grpc_servicer: GatewayMock):
        grpc_servicer.active_jobs[job_from_task.key] = job_from_task

        await job_poller.poll_once()

        job: Job = queue.get_nowait()
        assert job.key == job_from_task.key

    async def test_no_job_is_polled(self, job_poller: JobPoller, queue: asyncio.Queue):
        await job_poller.poll_once()

        assert queue.empty()

    async def test_job_is_added_to_task_state(self, job_poller: JobPoller, job_from_task: Job, grpc_servicer: GatewayMock):
        grpc_servicer.active_jobs[job_from_task.key] = job_from_task

        await job_poller.poll_once()

        assert job_poller.task_state.count_active() == 1


class TestShouldPoll:
    def test_should_poll_returns_expected_result_when_disconnected(self, job_poller: JobPoller):
        job_poller.zeebe_adapter.connected = False
        job_poller.zeebe_adapter.retrying_connection = False

        assert not job_poller.should_poll()

    def test_continues_polling_when_retrying_connection(self, job_poller: JobPoller):
        job_poller.zeebe_adapter.connected = False
        job_poller.zeebe_adapter.retrying_connection = True

        assert job_poller.should_poll()

    @pytest.mark.asyncio
    async def test_stops_polling_after_poller_is_stopped(self, job_poller: JobPoller):
        await job_poller.stop()

        assert not job_poller.should_poll()

    def test_stops_polling_after_max_tasks_is_reached(self, job_poller: JobPoller):
        job_poller.max_task_count = 0

        assert not job_poller.should_poll()


class TestMaxJobsToActivate:
    def test_returns_smallest_option(self, job_poller: JobPoller):
        job_poller.max_task_count = 0

        max_jobs_to_activate = job_poller.calculate_max_jobs_to_activate()

        assert max_jobs_to_activate == 0

    def test_returns_zero_when_max_number_of_jobs_are_running(self, job_poller: JobPoller):
        for _ in range(job_poller.max_task_count):
            job = random_job()
            job_poller.task_state.add(job)

        max_jobs_to_activate = job_poller.calculate_max_jobs_to_activate()

        assert max_jobs_to_activate == 0

    calculate_max_jobs_to_activate_cases = dict(
        max_task_count_minus_active_decides=(4, 10, 12, 6),
        max_task_count_minus_active_decides_2=(4, 12, 10, 8),
        max_task_count_minus_active_decides_zero_free=(4, 4, 12, 0),
        max_jobs_to_activate_decides=(4, 10, 5, 5),
        max_jobs_to_activate_decides_zero_active=(0, 10, 5, 5)
    )

    @pytest.mark.parametrize("active_jobs,max_worker_task_count,max_jobs_to_activate_on_task,expected",
                             calculate_max_jobs_to_activate_cases.values(),
                             ids=calculate_max_jobs_to_activate_cases.keys())
    def test_calculate_max_jobs_to_activate(self, job_poller: JobPoller, active_jobs: int, max_worker_task_count: int,
                                            max_jobs_to_activate_on_task: int, expected: int):
        job_poller.max_task_count = max_worker_task_count
        job_poller.task.config.max_jobs_to_activate = max_jobs_to_activate_on_task

        for _ in range(active_jobs):
            job = random_job()
            job_poller.task_state.add(job)

        max_jobs_to_activate = job_poller.calculate_max_jobs_to_activate()

        assert max_jobs_to_activate == expected



from typing import Awaitable, Callable, Union

from pyzeebe import Job

DecoratorRunner = Callable[[Job], Awaitable[Job]]
JobHandler = Callable[[Job], Awaitable[Job]]

SyncTaskDecorator = Callable[[Job], Job]
AsyncTaskDecorator = Callable[[Job], Awaitable[Job]]
TaskDecorator = Union[SyncTaskDecorator, AsyncTaskDecorator]



from random import randint
from uuid import uuid4

from pyzeebe.grpc_internals.zeebe_adapter import ZeebeAdapter
from pyzeebe.job.job import Job
from pyzeebe.task import task_builder
from pyzeebe.task.task import Task
from pyzeebe.task.task_config import TaskConfig

RANDOM_RANGE = 1000000000


def random_job(task: Task = task_builder.build_task(lambda x: {"x": x}, TaskConfig("test", lambda: None, 10000, 32, [], False, "", [], [])),
               zeebe_adapter: ZeebeAdapter = None) -> Job:
    return Job(_type=task.type, key=randint(0, RANDOM_RANGE), worker=str(uuid4()),
               retries=randint(0, 10), process_instance_key=randint(0, RANDOM_RANGE),
               bpmn_process_id=str(uuid4()), process_definition_version=randint(0, 100),
               process_definition_key=randint(0, RANDOM_RANGE), element_id=str(uuid4()),
               element_instance_key=randint(0, RANDOM_RANGE), variables={}, custom_headers={},
               deadline=randint(0, RANDOM_RANGE), zeebe_adapter=zeebe_adapter)



from typing import List
from uuid import uuid4

import grpc
import pytest

from pyzeebe import TaskDecorator, ZeebeTaskRouter
from pyzeebe.errors import DuplicateTaskTypeError
from pyzeebe.job.job import Job
from pyzeebe.task.task import Task
from pyzeebe.worker.worker import ZeebeWorker


class TestAddTask:
    def test_add_task(self, zeebe_worker: ZeebeWorker, task: Task):
        zeebe_worker._add_task(task)

        assert zeebe_worker.get_task(task.type) == task

    def test_raises_on_duplicate(self, zeebe_worker: ZeebeWorker, task: Task):
        zeebe_worker._add_task(task)
        with pytest.raises(DuplicateTaskTypeError):
            zeebe_worker._add_task(task)

    def test_only_one_task_added(self, zeebe_worker: ZeebeWorker):
        @zeebe_worker.task(str(uuid4()))
        def dummy_function():
            pass

        assert len(zeebe_worker.tasks) == 1

    def test_task_type_saved(self, zeebe_worker: ZeebeWorker, task: Task):
        zeebe_worker._add_task(task)

        assert zeebe_worker.get_task(task.type).type == task.type

    def test_variables_to_fetch_match_function_parameters(self, zeebe_worker: ZeebeWorker, task_type: str):
        expected_variables_to_fetch = ["x"]

        @zeebe_worker.task(task_type)
        def dummy_function(x):
            pass

        assert zeebe_worker.get_task(
            task_type).config.variables_to_fetch == expected_variables_to_fetch


class TestDecorator:
    def test_add_before_decorator(self, zeebe_worker: ZeebeWorker, decorator: TaskDecorator):
        zeebe_worker.before(decorator)
        assert len(zeebe_worker._before) == 1
        assert decorator in zeebe_worker._before

    def test_add_after_decorator(self, zeebe_worker: ZeebeWorker, decorator: TaskDecorator):
        zeebe_worker.after(decorator)
        assert len(zeebe_worker._after) == 1
        assert decorator in zeebe_worker._after

    def test_add_constructor_before_decorator(
        self, aio_grpc_channel: grpc.aio.Channel, decorator: TaskDecorator
    ):
        zeebe_worker = ZeebeWorker(aio_grpc_channel, before=[decorator])
        assert len(zeebe_worker._before) == 1
        assert decorator in zeebe_worker._before

    def test_add_constructor_after_decorator(
        self, aio_grpc_channel: grpc.aio.Channel, decorator: TaskDecorator
    ):
        zeebe_worker = ZeebeWorker(aio_grpc_channel, after=[decorator])
        assert len(zeebe_worker._after) == 1
        assert decorator in zeebe_worker._after


class TestIncludeRouter:
    def test_include_router_adds_task(self, zeebe_worker: ZeebeWorker, router: ZeebeTaskRouter, task_type: str):
        self.include_router_with_task(zeebe_worker, router, task_type)

        assert zeebe_worker.get_task(task_type) is not None

    def test_include_multiple_routers(self, zeebe_worker: ZeebeWorker, routers: List[ZeebeTaskRouter]):
        for router in routers:
            self.include_router_with_task(zeebe_worker, router)

        assert len(zeebe_worker.tasks) == len(routers)

    @pytest.mark.asyncio
    async def test_router_before_decorator(self, zeebe_worker: ZeebeWorker, router: ZeebeTaskRouter,
                                           decorator: TaskDecorator,
                                           mocked_job_with_adapter: Job):
        router.before(decorator)
        task = self.include_router_with_task(zeebe_worker, router)

        await task.job_handler(mocked_job_with_adapter)

        decorator.assert_called_once()

    @pytest.mark.asyncio
    async def test_router_after_decorator(self, zeebe_worker: ZeebeWorker, router: ZeebeTaskRouter,
                                          decorator: TaskDecorator,
                                          mocked_job_with_adapter: Job):
        router.after(decorator)
        task = self.include_router_with_task(zeebe_worker, router)

        await task.job_handler(mocked_job_with_adapter)

        decorator.assert_called_once()

    @staticmethod
    def include_router_with_task(zeebe_worker: ZeebeWorker, router: ZeebeTaskRouter, task_type: str = None) -> Task:
        task_type = task_type or str(uuid4())

        @router.task(task_type)
        def dummy_function():
            return {}

        zeebe_worker.include_router(router)
        return zeebe_worker.get_task(task_type)



import logging
import os
from typing import Optional

import grpc
from zeebe_grpc.gateway_pb2_grpc import GatewayStub

from pyzeebe.errors import (ZeebeBackPressureError,
                            ZeebeGatewayUnavailableError, ZeebeInternalError)

logger = logging.getLogger(__name__)


class ZeebeAdapterBase:
    def __init__(
        self, grpc_channel: grpc.aio.Channel, max_connection_retries: int = -1
    ):
        self._channel = grpc_channel
        self._gateway_stub = GatewayStub(grpc_channel)
        self.connected = True
        self.retrying_connection = False
        self._max_connection_retries = max_connection_retries
        self._current_connection_retries = 0

    def _should_retry(self):
        return (
            self._max_connection_retries == -1
            or self._current_connection_retries < self._max_connection_retries
        )

    async def _common_zeebe_grpc_errors(self, rpc_error: grpc.aio.AioRpcError):
        if self.is_error_status(rpc_error, grpc.StatusCode.RESOURCE_EXHAUSTED):
            raise ZeebeBackPressureError()
        elif self.is_error_status(rpc_error, grpc.StatusCode.UNAVAILABLE):
            self._current_connection_retries += 1
            if not self._should_retry():
                await self._close()
            raise ZeebeGatewayUnavailableError()
        elif self.is_error_status(rpc_error, grpc.StatusCode.INTERNAL):
            self._current_connection_retries += 1
            if not self._should_retry():
                await self._close()
            raise ZeebeInternalError()
        else:
            raise rpc_error

    @staticmethod
    def is_error_status(rpc_error: grpc.aio.AioRpcError, status_code: grpc.StatusCode):
        return rpc_error.code() == status_code

    async def _close(self):
        try:
            await self._channel.close()
        except Exception as exception:
            logger.exception(
                f"Failed to close channel, {type(exception).__name__} exception was raised"
            )



import inspect
from typing import Callable, List

import pytest

from pyzeebe.function_tools import async_tools
from tests.unit.utils.function_tools import functions_are_all_async


class TestAsyncify:
    def test_returns_async_function(self):
        async_function = async_tools.asyncify(lambda x: x)

        assert inspect.iscoroutinefunction(async_function)

    @pytest.mark.asyncio
    async def test_returned_function_returns_expected_value(self):
        expected_result = 5
        async_function = async_tools.asyncify(lambda: expected_result)

        assert await async_function() == expected_result

    @pytest.mark.asyncio
    async def test_returned_function_accepts_keyword_arguments(self):
        async_function = async_tools.asyncify(lambda x, y, z: x + y + z)

        assert await async_function(x=1, y=1, z=1) == 3


class TestAsyncifyAllFunctions:
    def sync_function(self):
        return

    async def async_function(self):
        return

    def test_changes_sync_function(self):
        functions = [self.sync_function]

        async_functions = async_tools.asyncify_all_functions(functions)

        assert functions_are_all_async(async_functions)

    def test_async_function_remains_unchanged(self):
        functions = [self.async_function]

        async_functions = async_tools.asyncify_all_functions(functions)

        assert async_functions[0] == functions[0]


class TestIsAsyncFunction:
    def test_with_normal_function(self):
        def normal_function():
            return 1

        assert not async_tools.is_async_function(normal_function)

    def test_with_async_function(self):
        async def async_function():
            return 1

        assert async_tools.is_async_function(async_function)



from uuid import uuid4

import grpc
import pytest
import responses
from mock import Mock, patch
from requests import HTTPError

from pyzeebe import create_camunda_cloud_channel
from pyzeebe.errors import (InvalidCamundaCloudCredentialsError,
                            InvalidOAuthCredentialsError)


@pytest.fixture
def mocked_responses() -> responses.RequestsMock:
    with responses.RequestsMock() as response_mock:
        yield response_mock


@pytest.fixture
def client_id() -> str:
    return str(uuid4())


@pytest.fixture
def client_secret() -> str:
    return str(uuid4())


@pytest.fixture
def cluster_id() -> str:
    return str(uuid4())


@pytest.fixture
def url() -> str:
    return "https://login.cloud.camunda.io/oauth/token"


@pytest.fixture
def access_token() -> str:
    return str(uuid4())


@pytest.fixture
def mock_access_token_response(
    mocked_responses: responses.RequestsMock, url: str, access_token: str
):
    mocked_responses.add(
        responses.POST, url, json={"access_token": access_token}, status=200
    )


class TestCamundaCloudChannel:
    @pytest.fixture(autouse=True)
    def secure_channel_mock(self, aio_grpc_channel: grpc.aio.Channel) -> Mock:
        with patch("grpc.aio.secure_channel", return_value=aio_grpc_channel) as mock:
            yield mock

    @pytest.mark.usefixtures("mock_access_token_response")
    def test_returns_grpc_channel(
        self,
        mocked_responses: responses.RequestsMock,
        client_id: str,
        client_secret: str,
        cluster_id: str,
    ):
        channel = create_camunda_cloud_channel(client_id, client_secret, cluster_id)

        assert isinstance(channel, grpc.aio.Channel)

    @pytest.mark.usefixtures("mock_access_token_response")
    def test_gets_access_token_from_camunda_cloud(
        self,
        mocked_responses: responses.RequestsMock,
        client_id: str,
        client_secret: str,
        cluster_id: str,
    ):
        create_camunda_cloud_channel(client_id, client_secret, cluster_id)

        assert len(mocked_responses.calls) == 1

    @pytest.mark.usefixtures("mock_access_token_response")
    def test_gets_access_token_using_correct_parameters(
        self,
        mocked_responses: responses.RequestsMock,
        client_id: str,
        client_secret: str,
        cluster_id: str,
    ):
        expected_request_body = (
            f"client_id={client_id}&client_secret={client_secret}&audience={cluster_id}"
        )

        create_camunda_cloud_channel(client_id, client_secret, cluster_id)

        request = mocked_responses.calls[0].request
        assert request.body == expected_request_body

    def test_raises_on_invalid_credentials(
        self,
        mocked_responses: responses.RequestsMock,
        url: str,
        client_id: str,
        client_secret: str,
        cluster_id: str,
    ):
        mocked_responses.add(
            responses.POST, url, status=400
        )  # Camunda cloud returns 400 when invalid credentials are provided

        with pytest.raises(InvalidCamundaCloudCredentialsError):
            create_camunda_cloud_channel(client_id, client_secret, cluster_id)

    @pytest.mark.usefixtures("mock_access_token_response")
    def test_creates_channel_using_grpc_credentials(
        self,
        client_id: str,
        client_secret: str,
        cluster_id: str,
        secure_channel_mock: Mock,
    ):
        create_camunda_cloud_channel(client_id, client_secret, cluster_id)

        secure_channel_call = secure_channel_mock.mock_calls[0]
        arguments = [arg for arg in secure_channel_call.args]
        assert any(
            isinstance(arg, grpc.ChannelCredentials) for arg in arguments
        ), "None of the arguments to grpc.aio.create_secure_channel were of type grpc.ChannelCredentials"



from pyzeebe.job.job import Job


def no_param():
    pass


def one_param(x):
    pass


def multiple_params(x, y, z):
    pass


def one_keyword_param(x=0):
    pass


def multiple_keyword_param(x=0, y=0, z=0):
    pass


def positional_and_keyword_params(x, y=0):
    pass


def args_param(*args):
    pass


def kwargs_param(**kwargs):
    pass


def standard_named_params(args, kwargs):
    pass


def with_job_parameter(job: Job):
    pass


def with_multiple_job_parameters(job: Job, job2: Job):
    pass


def lambda_no_params(): return None
def lambda_one_param(x): return None
def lambda_multiple_params(x, y, z): return None
def lambda_one_keyword_param(x=0): return None
def lambda_multiple_keyword_params(x=0, y=0, z=0): return None
def lambda_positional_and_keyword_params(x, y=0): return None



import logging
from typing import Callable, List, Optional, Tuple

from pyzeebe.errors import (BusinessError, DuplicateTaskTypeError,
                            NoVariableNameGivenError, TaskNotFoundError)
from pyzeebe.function_tools import parameter_tools
from pyzeebe.job.job import Job
from pyzeebe.task import task_builder
from pyzeebe.task.exception_handler import ExceptionHandler
from pyzeebe.task.task import Task
from pyzeebe.task.task_config import TaskConfig
from pyzeebe.task.types import TaskDecorator

logger = logging.getLogger(__name__)


async def default_exception_handler(e: Exception, job: Job) -> None:
    logger.warning(f"Task type: {job.type} - failed job {job}. Error: {e}.")
    if isinstance(e, BusinessError):
        await job.set_error_status(f"Failed job. Recoverable error: {e}", error_code=e.error_code)
    else:
        await job.set_failure_status(f"Failed job. Error: {e}")


class ZeebeTaskRouter:
    def __init__(self, before: List[TaskDecorator] = None, after: List[TaskDecorator] = None):
        """
        Args:
            before (List[TaskDecorator]): Decorators to be performed before each task
            after (List[TaskDecorator]): Decorators to be performed after each task
        """
        self._before: List[TaskDecorator] = before or []
        self._after: List[TaskDecorator] = after or []
        self.tasks: List[Task] = []

    def task(self, task_type: str, exception_handler: ExceptionHandler = default_exception_handler,
             variables_to_fetch: Optional[List[str]] = None, timeout_ms: int = 10000, max_jobs_to_activate: int = 32,
             before: List[TaskDecorator] = None, after: List[TaskDecorator] = None, single_value: bool = False,
             variable_name: str = None):
        """
        Decorator to create a task

        Args:
            task_type (str): The task type
            exception_handler (ExceptionHandler): Handler that will be called when a job fails. 
            variables_to_fetch (Optional[List[str]]): The variables to request from Zeebe when activating jobs.
            timeout_ms (int): Maximum duration of the task in milliseconds. If the timeout is surpassed Zeebe will give up 
                                on the worker and retry it. Default: 10000 (10 seconds).
            max_jobs_to_activate (int):  Maximum jobs the worker will execute in parallel (of this task). Default: 32
            before (List[TaskDecorator]): All decorators which should be performed before the task.
            after (List[TaskDecorator]): All decorators which should be performed after the task.
            single_value (bool): If the function returns a single value (int, string, list) and not a dictionary set
                                 this to True. Default: False
            variable_name (str): If single_value then this will be the variable name given to zeebe:
                                        { <variable_name>: <function_return_value> }

        Raises:
            DuplicateTaskTypeError: If a task from the router already exists in the worker
            NoVariableNameGivenError: When single_value is set, but no variable_name is given
        """
        def task_wrapper(task_function: Callable):
            config = TaskConfig(
                task_type,
                exception_handler,
                timeout_ms,
                max_jobs_to_activate,
                variables_to_fetch or parameter_tools.get_parameters_from_function(
                    task_function),
                single_value,
                variable_name or "",
                before or [],
                after or []
            )
            config_with_decorators = self._add_decorators_to_config(config)

            task = task_builder.build_task(
                task_function, config_with_decorators
            )
            self._add_task(task)
            return task_function

        return task_wrapper

    def _add_task(self, task: Task):
        self._is_task_duplicate(task.type)
        self.tasks.append(task)

    def _add_decorators_to_config(self, config: TaskConfig) -> TaskConfig:
        new_task_config = TaskConfig(
            type=config.type,
            exception_handler=config.exception_handler,
            timeout_ms=config.timeout_ms,
            max_jobs_to_activate=config.max_jobs_to_activate,
            variables_to_fetch=config.variables_to_fetch,
            single_value=config.single_value,
            variable_name=config.variable_name,
            before=self._before + config.before,  # type: ignore
            after=config.after + self._after  # type: ignore
        )
        return new_task_config

    def _is_task_duplicate(self, task_type: str) -> None:
        try:
            self.get_task(task_type)
            raise DuplicateTaskTypeError(task_type)
        except TaskNotFoundError:
            return

    def before(self, *decorators: TaskDecorator) -> None:
        """
        Add decorators to be performed before a job is run

        Args:
            decorators (Iterable[TaskDecorator]): The decorators to be performed before each job is run
        """
        self._before.extend(decorators)

    def after(self, *decorators: TaskDecorator) -> None:
        """
        Add decorators to be performed after a job is run

        Args:
            decorators (Iterable[TaskDecorator]): The decorators to be performed after each job is run
        """
        self._after.extend(decorators)

    def remove_task(self, task_type: str) -> Task:
        """
        Remove a task

        Args:
            task_type (str): The type of the wanted task

        Returns:
            Task: The task that was removed

        Raises:
             TaskNotFoundError: If no task with specified type exists

        """
        task_index = self._get_task_index(task_type)
        return self.tasks.pop(task_index)

    def get_task(self, task_type: str) -> Task:
        """
        Get a task by its type

        Args:
            task_type (str): The type of the wanted task

        Returns:
            Task: The wanted task

        Raises:
             TaskNotFoundError: If no task with specified type exists

        """
        return self._get_task_and_index(task_type)[0]

    def _get_task_index(self, task_type: str) -> int:
        return self._get_task_and_index(task_type)[1]

    def _get_task_and_index(self, task_type: str) -> Tuple[Task, int]:
        for index, task in enumerate(self.tasks):
            if task.type == task_type:
                return task, index
        raise TaskNotFoundError(f"Could not find task {task_type}")



import json
from typing import Dict

import grpc
from zeebe_grpc.gateway_pb2 import (PublishMessageRequest,
                                    PublishMessageResponse)

from pyzeebe.errors import MessageAlreadyExistsError
from pyzeebe.grpc_internals.zeebe_adapter_base import ZeebeAdapterBase


class ZeebeMessageAdapter(ZeebeAdapterBase):
    async def publish_message(self, name: str, correlation_key: str, time_to_live_in_milliseconds: int,
                              variables: Dict, message_id: str = None) -> PublishMessageResponse:
        try:
            return await self._gateway_stub.PublishMessage(
                PublishMessageRequest(name=name, correlationKey=correlation_key, messageId=message_id,
                                      timeToLive=time_to_live_in_milliseconds, variables=json.dumps(variables)))
        except grpc.aio.AioRpcError as rpc_error:
            if self.is_error_status(rpc_error, grpc.StatusCode.ALREADY_EXISTS):
                raise MessageAlreadyExistsError()
            else:
                await self._common_zeebe_grpc_errors(rpc_error)



import json
from random import randint
from typing import Dict, List
from uuid import uuid4

import grpc
from mock import patch
from zeebe_grpc.gateway_pb2 import *
from zeebe_grpc.gateway_pb2_grpc import GatewayServicer

from pyzeebe.job.job import Job
from pyzeebe.job.job_status import JobStatus
from pyzeebe.task.task import Task
from tests.unit.utils.random_utils import RANDOM_RANGE, random_job


@patch("grpc.insecure_channel")
def mock_channel():
    pass


class GatewayMock(GatewayServicer):
    # TODO: Mock behaviour of zeebe

    def __init__(self):
        self.deployed_processes = {}
        self.active_processes = {}
        self.active_jobs: Dict[int, Job] = {}
        self.messages = {}

    def ActivateJobs(self, request, context):
        if not request.type:
            context.set_code(grpc.StatusCode.INVALID_ARGUMENT)
            return ActivateJobsResponse()

        if request.maxJobsToActivate <= 0:
            context.set_code(grpc.StatusCode.INVALID_ARGUMENT)
            return ActivateJobsResponse()

        if request.timeout <= 0:
            context.set_code(grpc.StatusCode.INVALID_ARGUMENT)
            return ActivateJobsResponse()

        if not request.worker:
            context.set_code(grpc.StatusCode.INVALID_ARGUMENT)
            return ActivateJobsResponse()

        jobs = []
        for active_job in self.active_jobs.values():
            if active_job.type == request.type:
                jobs.append(ActivatedJob(key=active_job.key, type=active_job.type,
                                         processInstanceKey=active_job.process_instance_key,
                                         bpmnProcessId=active_job.bpmn_process_id,
                                         processDefinitionVersion=active_job.process_definition_version,
                                         processDefinitionKey=active_job.process_definition_key,
                                         elementId=active_job.element_id,
                                         elementInstanceKey=active_job.element_instance_key,
                                         customHeaders=json.dumps(
                                             active_job.custom_headers),
                                         worker=active_job.worker, retries=active_job.retries,
                                         deadline=active_job.deadline,
                                         variables=json.dumps(active_job.variables)))
        yield ActivateJobsResponse(jobs=jobs)

    def CompleteJob(self, request, context):
        if request.jobKey in self.active_jobs.keys():
            active_job = self.active_jobs.get(request.jobKey)
            self.handle_job(active_job, JobStatus.Completed, context)
            return CompleteJobResponse()
        else:
            context.set_code(grpc.StatusCode.NOT_FOUND)
            return CompleteJobResponse()

    def FailJob(self, request, context):
        if request.jobKey in self.active_jobs.keys():
            active_job = self.active_jobs.get(request.jobKey)
            self.handle_job(active_job, JobStatus.Failed, context)
            return FailJobResponse()
        else:
            context.set_code(grpc.StatusCode.NOT_FOUND)
            return FailJobResponse()

    def ThrowError(self, request, context):
        if request.jobKey in self.active_jobs.keys():
            active_job = self.active_jobs.get(request.jobKey)
            self.handle_job(active_job, JobStatus.ErrorThrown, context)
            return CompleteJobResponse()
        else:
            context.set_code(grpc.StatusCode.NOT_FOUND)
            return CompleteJobResponse()

    @staticmethod
    def handle_job(job: Job, status_on_deactivate: JobStatus, context):
        if job.status != JobStatus.Running:
            context.set_code(grpc.StatusCode.FAILED_PRECONDITION)
        else:
            job.status = status_on_deactivate
        return context

    def CreateProcessInstance(self, request, context):
        if request.bpmnProcessId in self.deployed_processes.keys():
            for task in self.deployed_processes[request.bpmnProcessId]["tasks"]:
                job = random_job(task)
                self.active_jobs[job.key] = job

            process_instance_key = randint(0, RANDOM_RANGE)
            self.active_processes[process_instance_key] = request.bpmnProcessId
            return CreateProcessInstanceResponse(processDefinitionKey=randint(0, RANDOM_RANGE),
                                                 bpmnProcessId=request.bpmnProcessId,
                                                 version=request.version, processInstanceKey=process_instance_key)
        else:
            context.set_code(grpc.StatusCode.NOT_FOUND)
            return CreateProcessInstanceResponse()

    def CreateProcessInstanceWithResult(self, request, context):
        if request.request.bpmnProcessId in self.deployed_processes.keys():
            process_instance_key = randint(0, RANDOM_RANGE)
            self.active_processes[process_instance_key] = request.request.bpmnProcessId

            return CreateProcessInstanceWithResultResponse(processDefinitionKey=request.request.processDefinitionKey,
                                                           processInstanceKey=process_instance_key,
                                                           bpmnProcessId=request.request.bpmnProcessId,
                                                           version=randint(0, 10), variables=request.request.variables)
        else:
            context.set_code(grpc.StatusCode.NOT_FOUND)
            return CreateProcessInstanceWithResultResponse()

    def CancelProcessInstance(self, request, context):
        if request.processInstanceKey in self.active_processes.keys():
            del self.active_processes[request.processInstanceKey]
            return CancelProcessInstanceResponse()
        else:
            context.set_code(grpc.StatusCode.NOT_FOUND)
            return CancelProcessInstanceResponse()

    def DeployProcess(self, request, context):
        processes = []
        for process in request.processes:
            process_metadata = ProcessMetadata(bpmnProcessId=str(uuid4()), version=randint(0, 10),
                                               processDefinitionKey=randint(0, RANDOM_RANGE), resourceName=process.name)
            processes.append(process_metadata)

        return DeployProcessResponse(key=randint(0, RANDOM_RANGE), processes=processes)

    def PublishMessage(self, request, context):
        if request.messageId in self.messages.keys():
            context.set_code(grpc.StatusCode.ALREADY_EXISTS)
        else:
            self.messages[request.messageId] = request.correlationKey
        return PublishMessageResponse()

    def mock_deploy_process(self, bpmn_process_id: str, version: int, tasks: List[Task]):
        self.deployed_processes[bpmn_process_id] = {"bpmn_process_id": bpmn_process_id, "version": version,
                                                    "tasks": tasks}



import functools

from pyzeebe.function_tools import AsyncFunction, DictFunction


def convert_to_dict_function(single_value_function: AsyncFunction, variable_name: str) -> DictFunction:
    @functools.wraps(single_value_function)
    async def inner_fn(*args, **kwargs):
        return {variable_name: await single_value_function(*args, **kwargs)}

    return inner_fn



import json
import logging
from typing import AsyncGenerator, Dict, List

import grpc
from zeebe_grpc.gateway_pb2 import (ActivateJobsRequest, CompleteJobRequest,
                                    CompleteJobResponse, FailJobRequest,
                                    FailJobResponse, ThrowErrorRequest,
                                    ThrowErrorResponse)

from pyzeebe.errors import (ActivateJobsRequestInvalidError,
                            JobAlreadyDeactivatedError, JobNotFoundError)
from pyzeebe.grpc_internals.zeebe_adapter_base import ZeebeAdapterBase
from pyzeebe.job.job import Job

logger = logging.getLogger(__name__)


class ZeebeJobAdapter(ZeebeAdapterBase):
    async def activate_jobs(self, task_type: str, worker: str, timeout: int, max_jobs_to_activate: int,
                            variables_to_fetch: List[str], request_timeout: int) -> AsyncGenerator[Job, None]:
        try:
            async for response in self._gateway_stub.ActivateJobs(
                    ActivateJobsRequest(type=task_type, worker=worker, timeout=timeout,
                                        maxJobsToActivate=max_jobs_to_activate,
                                        fetchVariable=variables_to_fetch, requestTimeout=request_timeout)):
                for raw_job in response.jobs:
                    job = self._create_job_from_raw_job(raw_job)
                    logger.debug(f"Got job: {job} from zeebe")
                    yield job
        except grpc.aio.AioRpcError as rpc_error:
            if self.is_error_status(rpc_error, grpc.StatusCode.INVALID_ARGUMENT):
                raise ActivateJobsRequestInvalidError(
                    task_type, worker, timeout, max_jobs_to_activate)
            else:
                await self._common_zeebe_grpc_errors(rpc_error)

    def _create_job_from_raw_job(self, response) -> Job:
        return Job(key=response.key, _type=response.type,
                   process_instance_key=response.processInstanceKey,
                   bpmn_process_id=response.bpmnProcessId,
                   process_definition_version=response.processDefinitionVersion,
                   process_definition_key=response.processDefinitionKey,
                   element_id=response.elementId,
                   element_instance_key=response.elementInstanceKey,
                   custom_headers=json.loads(response.customHeaders),
                   worker=response.worker,
                   retries=response.retries,
                   deadline=response.deadline,
                   variables=json.loads(response.variables),
                   zeebe_adapter=self)

    async def complete_job(self, job_key: int, variables: Dict) -> CompleteJobResponse:
        try:
            return await self._gateway_stub.CompleteJob(CompleteJobRequest(jobKey=job_key, variables=json.dumps(variables)))
        except grpc.aio.AioRpcError as rpc_error:
            if self.is_error_status(rpc_error, grpc.StatusCode.NOT_FOUND):
                raise JobNotFoundError(job_key=job_key)
            elif self.is_error_status(rpc_error, grpc.StatusCode.FAILED_PRECONDITION):
                raise JobAlreadyDeactivatedError(job_key=job_key)
            else:
                await self._common_zeebe_grpc_errors(rpc_error)

    async def fail_job(self, job_key: int, retries: int, message: str) -> FailJobResponse:
        try:
            return await self._gateway_stub.FailJob(FailJobRequest(jobKey=job_key, retries=retries, errorMessage=message))
        except grpc.aio.AioRpcError as rpc_error:
            if self.is_error_status(rpc_error, grpc.StatusCode.NOT_FOUND):
                raise JobNotFoundError(job_key=job_key)
            elif self.is_error_status(rpc_error, grpc.StatusCode.FAILED_PRECONDITION):
                raise JobAlreadyDeactivatedError(job_key=job_key)
            else:
                await self._common_zeebe_grpc_errors(rpc_error)

    async def throw_error(self, job_key: int, message: str, error_code: str = "") -> ThrowErrorResponse:
        try:
            return await self._gateway_stub.ThrowError(
                ThrowErrorRequest(jobKey=job_key, errorMessage=message, errorCode=error_code))
        except grpc.aio.AioRpcError as rpc_error:
            if self.is_error_status(rpc_error, grpc.StatusCode.NOT_FOUND):
                raise JobNotFoundError(job_key=job_key)
            elif self.is_error_status(rpc_error, grpc.StatusCode.FAILED_PRECONDITION):
                raise JobAlreadyDeactivatedError(job_key=job_key)
            else:
                await self._common_zeebe_grpc_errors(rpc_error)



from random import randint
from threading import Event
from uuid import uuid4

import grpc
import pytest
from mock import AsyncMock, MagicMock, patch

from pyzeebe import Job, ZeebeClient, ZeebeWorker
from pyzeebe.grpc_internals.zeebe_adapter import ZeebeAdapter
from pyzeebe.task import task_builder
from pyzeebe.task.task_config import TaskConfig
from pyzeebe.worker.task_router import ZeebeTaskRouter
from pyzeebe.worker.task_state import TaskState
from tests.unit.utils.gateway_mock import GatewayMock
from tests.unit.utils.random_utils import random_job


@pytest.fixture
def job_with_adapter(zeebe_adapter):
    return random_job(zeebe_adapter=zeebe_adapter)


@pytest.fixture
def mocked_job_with_adapter(job_with_adapter):
    job_with_adapter.set_success_status = AsyncMock()
    job_with_adapter.set_failure_status = AsyncMock()
    job_with_adapter.set_error_status = AsyncMock()
    return job_with_adapter


@pytest.fixture
def job_without_adapter():
    return random_job()


@pytest.fixture
def job_from_task(task):
    job = random_job(task)
    job.variables = dict(x=str(uuid4()))
    return job


@pytest.fixture
def zeebe_adapter(aio_grpc_channel: grpc.aio.Channel):
    adapter = ZeebeAdapter(aio_grpc_channel)
    return adapter


@pytest.fixture
def zeebe_client(aio_grpc_channel: grpc.aio.Channel):
    client = ZeebeClient(aio_grpc_channel)
    return client


@pytest.fixture
def zeebe_worker(aio_grpc_channel: grpc.aio.Channel):
    worker = ZeebeWorker(aio_grpc_channel)
    return worker


@pytest.fixture
def task(original_task_function, task_config):
    return task_builder.build_task(original_task_function, task_config)


@pytest.fixture
def first_active_job(task, job_from_task, grpc_servicer) -> str:
    grpc_servicer.active_jobs[job_from_task.key] = job_from_task
    return job_from_task


@pytest.fixture
def task_config(task_type):
    return TaskConfig(
        type=task_type,
        exception_handler=AsyncMock(),
        timeout_ms=10000,
        max_jobs_to_activate=32,
        variables_to_fetch=[],
        single_value=False,
        variable_name="",
        before=[],
        after=[],
    )


@pytest.fixture
def task_type():
    return str(uuid4())


@pytest.fixture
def original_task_function():
    def original_function():
        pass

    mock = MagicMock(wraps=original_function)
    mock.__code__ = original_function.__code__
    return mock


@pytest.fixture
def stop_after_test():
    stop_test = Event()
    yield stop_test
    stop_test.set()


@pytest.fixture
def handle_task_mock():
    with patch("pyzeebe.worker.worker.ZeebeWorker._handle_task") as mock:
        yield mock


@pytest.fixture
def stop_event_mock(zeebe_worker):
    with patch.object(zeebe_worker, "stop_event") as mock:
        yield mock


@pytest.fixture
def handle_not_alive_thread_spy(mocker):
    spy = mocker.spy(ZeebeWorker, "_handle_not_alive_thread")
    yield spy


@pytest.fixture
def router():
    return ZeebeTaskRouter()


@pytest.fixture
def routers():
    return [ZeebeTaskRouter() for _ in range(0, randint(2, 100))]


@pytest.fixture
def decorator():
    async def simple_decorator(job: Job) -> Job:
        return job

    return AsyncMock(wraps=simple_decorator)


@pytest.fixture(scope="module")
def grpc_add_to_server():
    from zeebe_grpc.gateway_pb2_grpc import add_GatewayServicer_to_server
    return add_GatewayServicer_to_server


@pytest.fixture(scope="module")
def grpc_servicer():
    return GatewayMock()


@pytest.fixture(scope="module")
def grpc_stub_cls(grpc_channel):
    from zeebe_grpc.gateway_pb2_grpc import GatewayStub
    return GatewayStub


@pytest.fixture
def aio_create_grpc_channel(request, grpc_addr, grpc_server):
    return grpc.aio.insecure_channel(grpc_addr)


@pytest.mark.asyncio
@pytest.fixture
async def aio_grpc_channel(aio_create_grpc_channel):
    async with aio_create_grpc_channel as channel:
        yield channel


@pytest.fixture
def task_state() -> TaskState:
    return TaskState()



import json
import os
from typing import Dict, Tuple

import aiofiles
import grpc
from zeebe_grpc.gateway_pb2 import (CancelProcessInstanceRequest,
                                    CreateProcessInstanceRequest,
                                    CreateProcessInstanceWithResultRequest,
                                    DeployProcessRequest,
                                    DeployProcessResponse,
                                    ProcessRequestObject)

from pyzeebe.errors import (InvalidJSONError,
                            ProcessDefinitionHasNoStartEventError,
                            ProcessDefinitionNotFoundError,
                            ProcessInstanceNotFoundError, ProcessInvalidError)
from pyzeebe.grpc_internals.zeebe_adapter_base import ZeebeAdapterBase


class ZeebeProcessAdapter(ZeebeAdapterBase):
    async def create_process_instance(self, bpmn_process_id: str, version: int, variables: Dict) -> int:
        try:
            response = await self._gateway_stub.CreateProcessInstance(
                CreateProcessInstanceRequest(bpmnProcessId=bpmn_process_id, version=version,
                                             variables=json.dumps(variables)))
        except grpc.aio.AioRpcError as rpc_error:
            await self._create_process_errors(rpc_error, bpmn_process_id, version, variables)
        return response.processInstanceKey

    async def create_process_instance_with_result(self, bpmn_process_id: str, version: int, variables: Dict,
                                                  timeout: int, variables_to_fetch) -> Tuple[int, Dict]:
        try:
            response = await self._gateway_stub.CreateProcessInstanceWithResult(
                CreateProcessInstanceWithResultRequest(
                    request=CreateProcessInstanceRequest(bpmnProcessId=bpmn_process_id, version=version,
                                                         variables=json.dumps(variables)),
                    requestTimeout=timeout, fetchVariables=variables_to_fetch))
        except grpc.aio.AioRpcError as rpc_error:
            await self._create_process_errors(rpc_error, bpmn_process_id, version, variables)
        return response.processInstanceKey, json.loads(response.variables)

    async def _create_process_errors(self, rpc_error: grpc.aio.AioRpcError, bpmn_process_id: str, version: int,
                                     variables: Dict) -> None:
        if self.is_error_status(rpc_error, grpc.StatusCode.NOT_FOUND):
            raise ProcessDefinitionNotFoundError(
                bpmn_process_id=bpmn_process_id, version=version)
        elif self.is_error_status(rpc_error, grpc.StatusCode.INVALID_ARGUMENT):
            raise InvalidJSONError(
                f"Cannot start process: {bpmn_process_id} with version {version}. Variables: {variables}")
        elif self.is_error_status(rpc_error, grpc.StatusCode.FAILED_PRECONDITION):
            raise ProcessDefinitionHasNoStartEventError(
                bpmn_process_id=bpmn_process_id)
        else:
            await self._common_zeebe_grpc_errors(rpc_error)

    async def cancel_process_instance(self, process_instance_key: int) -> None:
        try:
            await self._gateway_stub.CancelProcessInstance(
                CancelProcessInstanceRequest(processInstanceKey=process_instance_key))
        except grpc.aio.AioRpcError as rpc_error:
            if self.is_error_status(rpc_error, grpc.StatusCode.NOT_FOUND):
                raise ProcessInstanceNotFoundError(
                    process_instance_key=process_instance_key)
            else:
                await self._common_zeebe_grpc_errors(rpc_error)

    async def deploy_process(self, *process_file_path: str) -> DeployProcessResponse:
        try:
            return await self._gateway_stub.DeployProcess(
                DeployProcessRequest(processes=[await result for result in map(self._get_process_request_object,
                                                                               process_file_path)]))
        except grpc.aio.AioRpcError as rpc_error:
            if self.is_error_status(rpc_error, grpc.StatusCode.INVALID_ARGUMENT):
                raise ProcessInvalidError()
            else:
                await self._common_zeebe_grpc_errors(rpc_error)

    @staticmethod
    async def _get_process_request_object(process_file_path: str) -> ProcessRequestObject:
        async with aiofiles.open(process_file_path, "rb") as file:
            return ProcessRequestObject(name=os.path.basename(process_file_path),
                                        definition=await file.read())



import asyncio
import logging

from pyzeebe.errors import (ActivateJobsRequestInvalidError,
                            ZeebeBackPressureError,
                            ZeebeGatewayUnavailableError, ZeebeInternalError)
from pyzeebe.grpc_internals.zeebe_job_adapter import ZeebeJobAdapter
from pyzeebe.task.task import Task
from pyzeebe.worker.task_state import TaskState

logger = logging.getLogger(__name__)


class JobPoller:
    def __init__(self, zeebe_adapter: ZeebeJobAdapter, task: Task, queue: asyncio.Queue, worker_name: str,
                 request_timeout: int, task_state: TaskState, max_task_count: int):
        self.zeebe_adapter = zeebe_adapter
        self.task = task
        self.queue = queue
        self.worker_name = worker_name
        self.request_timeout = request_timeout
        self.task_state = task_state
        self.max_task_count = max_task_count
        self.stop_event = asyncio.Event()

    async def poll(self):
        while self.should_poll():
            await self.poll_once()

    async def poll_once(self):
        try:
            jobs = self.zeebe_adapter.activate_jobs(
                task_type=self.task.type,
                worker=self.worker_name,
                timeout=self.task.config.timeout_ms,
                max_jobs_to_activate=self.calculate_max_jobs_to_activate(),
                variables_to_fetch=self.task.config.variables_to_fetch,
                request_timeout=self.request_timeout,
            )
            async for job in jobs:
                self.task_state.add(job)
                await self.queue.put(job)
        except ActivateJobsRequestInvalidError:
            logger.warning(
                f"Activate job requests was invalid for task {self.task.type}"
            )
            raise
        except (ZeebeBackPressureError, ZeebeGatewayUnavailableError, ZeebeInternalError) as error:
            logger.warning(
                f"Failed to activate jobs from the gateway. Exception: {str(error)}. Retrying in 5 seconds..."
            )
            await asyncio.sleep(5)

    def should_poll(self) -> bool:
        return not self.stop_event.is_set() \
            and (self.zeebe_adapter.connected or self.zeebe_adapter.retrying_connection) \
            and self.calculate_max_jobs_to_activate() > 0

    def calculate_max_jobs_to_activate(self) -> int:
        worker_max_jobs = self.max_task_count - self.task_state.count_active()
        return min(worker_max_jobs, self.task.config.max_jobs_to_activate)

    async def stop(self):
        self.stop_event.set()
        await self.queue.join()



from typing import Dict, List, Tuple

import grpc

from pyzeebe.grpc_internals.zeebe_adapter import ZeebeAdapter


class ZeebeClient(object):
    """A zeebe client that can connect to a zeebe instance and perform actions."""

    def __init__(
        self, grpc_channel: grpc.aio.Channel, max_connection_retries: int = 10
    ):
        """
        Args:
            grpc_channel (grpc.aio.Channel): GRPC Channel connected to a Zeebe gateway
            max_connection_retries (int): Amount of connection retries before client gives up on connecting to zeebe. To setup with infinite retries use -1
        """

        self.zeebe_adapter = ZeebeAdapter(grpc_channel, max_connection_retries)

    async def run_process(
        self, bpmn_process_id: str, variables: Dict = None, version: int = -1
    ) -> int:
        """
        Run process

        Args:
            bpmn_process_id (str): The unique process id of the process.
            variables (dict): A dictionary containing all the starting variables the process needs. Must be JSONable.
            version (int): The version of the process. Default: -1 (latest)

        Returns:
            int: process_instance_key, the unique id of the running process generated by Zeebe.

        Raises:
            ProcessDefinitionNotFoundError: No process with bpmn_process_id exists
            InvalidJSONError: variables is not JSONable
            ProcessDefinitionHasNoStartEventError: The specified process does not have a start event
            ZeebeBackPressureError: If Zeebe is currently in back pressure (too many requests)
            ZeebeGatewayUnavailableError: If the Zeebe gateway is unavailable
            ZeebeInternalError: If Zeebe experiences an internal error

        """
        return await self.zeebe_adapter.create_process_instance(
            bpmn_process_id=bpmn_process_id, variables=variables or {}, version=version
        )

    async def run_process_with_result(
        self,
        bpmn_process_id: str,
        variables: Dict = None,
        version: int = -1,
        timeout: int = 0,
        variables_to_fetch: List[str] = None,
    ) -> Tuple[int, Dict]:
        """
        Run process and wait for the result.

        Args:
            bpmn_process_id (str): The unique process id of the process.
            variables (dict): A dictionary containing all the starting variables the process needs. Must be JSONable.
            version (int): The version of the process. Default: -1 (latest)
            timeout (int): How long to wait until a timeout occurs. Default: 0 (Zeebe default timeout)
            variables_to_fetch (List[str]): Which variables to get from the finished process

        Returns:
            tuple: (The process instance key, A dictionary of the end state of the process instance)

        Raises:
            ProcessDefinitionNotFoundError: No process with bpmn_process_id exists
            InvalidJSONError: variables is not JSONable
            ProcessDefinitionHasNoStartEventError: The specified process does not have a start event
            ZeebeBackPressureError: If Zeebe is currently in back pressure (too many requests)
            ZeebeGatewayUnavailableError: If the Zeebe gateway is unavailable
            ZeebeInternalError: If Zeebe experiences an internal error

        """
        return await self.zeebe_adapter.create_process_instance_with_result(
            bpmn_process_id=bpmn_process_id,
            variables=variables or {},
            version=version,
            timeout=timeout,
            variables_to_fetch=variables_to_fetch or [],
        )

    async def cancel_process_instance(self, process_instance_key: int) -> int:
        """
        Cancel a running process instance

        Args:
            process_instance_key (int): The key of the running process to cancel

        Returns:
            int: The process_instance_key

        Raises:
            ProcessInstanceNotFoundError: If no process instance with process_instance_key exists
            ZeebeBackPressureError: If Zeebe is currently in back pressure (too many requests)
            ZeebeGatewayUnavailableError: If the Zeebe gateway is unavailable
            ZeebeInternalError: If Zeebe experiences an internal error

        """
        await self.zeebe_adapter.cancel_process_instance(
            process_instance_key=process_instance_key
        )
        return process_instance_key

    async def deploy_process(self, *process_file_path: str) -> None:
        """
        Deploy one or more processes

        Args:
            process_file_path (str): The file path to a process definition file (bpmn/yaml)

        Raises:
            ProcessInvalidError: If one of the process file definitions is invalid
            ZeebeBackPressureError: If Zeebe is currently in back pressure (too many requests)
            ZeebeGatewayUnavailableError: If the Zeebe gateway is unavailable
            ZeebeInternalError: If Zeebe experiences an internal error

        """
        await self.zeebe_adapter.deploy_process(*process_file_path)

    async def publish_message(
        self,
        name: str,
        correlation_key: str,
        variables: Dict = None,
        time_to_live_in_milliseconds: int = 60000,
        message_id: str = None,
    ) -> None:
        """
        Publish a message

        Args:
            name (str): The message name
            correlation_key (str): The correlation key. For more info: https://docs.zeebe.io/glossary.html?highlight=correlation#correlation-key
            variables (dict): The variables the message should contain.
            time_to_live_in_milliseconds (int): How long this message should stay active. Default: 60000 ms (60 seconds)
            message_id (str): A unique message id. Useful for avoiding duplication. If a message with this id is still
                                active, a MessageAlreadyExists will be raised.

        Raises:
            MessageAlreadyExistError: If a message with message_id already exists
            ZeebeBackPressureError: If Zeebe is currently in back pressure (too many requests)
            ZeebeGatewayUnavailableError: If the Zeebe gateway is unavailable
            ZeebeInternalError: If Zeebe experiences an internal error

        """
        await self.zeebe_adapter.publish_message(
            name=name,
            correlation_key=correlation_key,
            time_to_live_in_milliseconds=time_to_live_in_milliseconds,
            variables=variables or {},
            message_id=message_id,
        )



from uuid import uuid4

import grpc
import pytest
from mock import Mock, patch

from pyzeebe import create_insecure_channel
from pyzeebe.channel.channel_options import get_channel_options
from pyzeebe.channel.utils import (DEFAULT_HOSTNAME, DEFAULT_PORT,
                                   create_address)


class TestCreateInsecureChannel:
    @pytest.fixture(autouse=True)
    def insecure_channel_mock(self, aio_grpc_channel: grpc.aio.Channel):
        with patch("grpc.aio.insecure_channel", return_value=aio_grpc_channel) as mock:
            yield mock

    def test_returns_aio_grpc_channel(self):
        channel = create_insecure_channel()

        assert isinstance(channel, grpc.aio.Channel)

    def test_calls_using_default_grpc_options(self, insecure_channel_mock: Mock):
        create_insecure_channel()

        insecure_channel_call = insecure_channel_mock.mock_calls[0]
        assert insecure_channel_call.kwargs["options"] == get_channel_options()

    def test_uses_default_address(self, insecure_channel_mock: Mock):
        create_insecure_channel()

        insecure_channel_call = insecure_channel_mock.mock_calls[0]
        assert insecure_channel_call.args[0] == create_address()

    def test_overrides_default_port_if_provided(self, insecure_channel_mock: Mock):
        port = 123

        create_insecure_channel(port=port)

        insecure_channel_call = insecure_channel_mock.mock_calls[0]
        assert insecure_channel_call.args[0] == f"{DEFAULT_HOSTNAME}:{port}"

    def test_overrides_default_hostname_if_provided(self, insecure_channel_mock: Mock):
        hostname = str(uuid4())

        create_insecure_channel(hostname=hostname)

        insecure_channel_call = insecure_channel_mock.mock_calls[0]
        assert insecure_channel_call.args[0] == f"{hostname}:{DEFAULT_PORT}"



from typing import Dict, Optional

import grpc

from pyzeebe.channel.channel_options import get_channel_options
from pyzeebe.channel.utils import create_address


def create_insecure_channel(
    hostname: Optional[str] = None,
    port: Optional[int] = None,
    channel_options: Optional[Dict] = None
) -> grpc.aio.Channel:
    """
    Create an insecure channel

    Args:
        hostname (Optional[str], optional): Zeebe gateway hostname
        port (Optional[int], optional): Zeebe gateway port
        channel_options (Optional[Dict], optional): GRPC channel options. See https://grpc.github.io/grpc/python/glossary.html#term-channel_arguments

    Returns:
        grpc.aio.Channel: A GRPC Channel connected to the Zeebe gateway.
    """
    address = create_address(hostname, port)
    return grpc.aio.insecure_channel(address, options=get_channel_options(channel_options))



import os
from typing import Optional

DEFAULT_HOSTNAME = "localhost"
DEFAULT_PORT = 26500
DEFAULT_ADDRESS = f"{DEFAULT_HOSTNAME}:{DEFAULT_PORT}"


def create_address(
    hostname: Optional[str] = None,
    port: Optional[int] = None,
) -> str:
    if hostname or port:
        return f"{hostname or DEFAULT_HOSTNAME}:{port or DEFAULT_PORT}"
    return os.getenv("ZEEBE_ADDRESS", DEFAULT_ADDRESS)



from uuid import uuid4

import mock
import pytest

from pyzeebe import TaskDecorator
from pyzeebe.errors import (BusinessError, DuplicateTaskTypeError,
                            TaskNotFoundError)
from pyzeebe.job.job import Job
from pyzeebe.task.task import Task
from pyzeebe.worker.task_router import (ZeebeTaskRouter,
                                        default_exception_handler)
from tests.unit.utils.random_utils import randint


def test_get_task(router: ZeebeTaskRouter, task: Task):
    router.tasks.append(task)

    found_task = router.get_task(task.type)

    assert found_task == task


def test_get_fake_task(router: ZeebeTaskRouter):
    with pytest.raises(TaskNotFoundError):
        router.get_task(str(uuid4()))


def test_get_task_index(router: ZeebeTaskRouter, task: Task):
    router.tasks.append(task)

    index = router._get_task_index(task.type)

    assert router.tasks[index] == task


def test_get_task_and_index(router: ZeebeTaskRouter, task: Task):
    router.tasks.append(task)

    found_task, index = router._get_task_and_index(task.type)

    assert router.tasks[index] == task
    assert found_task == task


def test_remove_task(router: ZeebeTaskRouter, task: Task):
    router.tasks.append(task)

    router.remove_task(task.type)

    assert task not in router.tasks


def test_remove_task_from_many(router: ZeebeTaskRouter, task: Task):
    router.tasks.append(task)

    for _ in range(1, randint(0, 100)):
        @router.task(str(uuid4()))
        def dummy_function():
            pass

    router.remove_task(task.type)

    assert task not in router.tasks


def test_remove_fake_task(router: ZeebeTaskRouter):
    with pytest.raises(TaskNotFoundError):
        router.remove_task(str(uuid4()))


def test_check_is_task_duplicate_with_duplicate(router: ZeebeTaskRouter, task: Task):
    router.tasks.append(task)
    with pytest.raises(DuplicateTaskTypeError):
        router._is_task_duplicate(task.type)


def test_no_duplicate_task_type_error_is_raised(router: ZeebeTaskRouter, task: Task):
    router._is_task_duplicate(task.type)


def test_add_before_decorator(router: ZeebeTaskRouter, decorator: TaskDecorator):
    router.before(decorator)

    assert len(router._before) == 1


def test_add_after_decorator(router: ZeebeTaskRouter, decorator: TaskDecorator):
    router.after(decorator)

    assert len(router._after) == 1


def test_add_before_decorator_through_constructor(decorator: TaskDecorator):
    router = ZeebeTaskRouter(before=[decorator])

    assert len(router._before) == 1


def test_add_after_decorator_through_constructor(decorator: TaskDecorator):
    router = ZeebeTaskRouter(after=[decorator])

    assert len(router._after) == 1


@pytest.mark.asyncio
async def test_default_exception_handler_logs_a_warning(mocked_job_with_adapter: Job):
    with mock.patch("pyzeebe.worker.task_router.logger.warning") as logging_mock:
        await default_exception_handler(Exception(), mocked_job_with_adapter)

        mocked_job_with_adapter.set_failure_status.assert_called()
        logging_mock.assert_called()


@pytest.mark.asyncio
async def test_default_exception_handler_uses_business_error(job_without_adapter):
    with mock.patch("pyzeebe.job.job.Job.set_error_status") as failure_mock:
        error_code = "custom-error-code"
        exception = BusinessError(error_code)
        await default_exception_handler(exception, job_without_adapter)
        failure_mock.assert_called_with(mock.ANY, error_code=error_code)


@pytest.mark.asyncio
async def test_default_exception_handler_warns_of_job_failure(job_without_adapter):
    with mock.patch("pyzeebe.worker.task_router.logger.warning") as logging_mock:
        with mock.patch("pyzeebe.job.job.Job.set_error_status"):
            exception = BusinessError("custom-error-code")
            await default_exception_handler(exception, job_without_adapter)
        logging_mock.assert_called()



import inspect
from typing import List, Optional

from pyzeebe.function_tools import Function
from pyzeebe.job.job import Job


def get_parameters_from_function(task_function: Function) -> List[str]:
    function_signature = inspect.signature(task_function)
    for _, parameter in function_signature.parameters.items():
        if parameter.kind in (inspect.Parameter.VAR_POSITIONAL, inspect.Parameter.VAR_KEYWORD):
            return []
    return list(function_signature.parameters)


def get_job_parameter_name(function: Function) -> Optional[str]:
    function_signature = inspect.signature(function)
    params = list(function_signature.parameters.values())
    for param in params:
        if param.annotation == Job:
            return param.name
    return None



from copy import deepcopy

import pytest

import pyzeebe.channel.channel_options
from pyzeebe.channel.channel_options import get_channel_options


def test_get_channel_options_returns_tuple_of_tuple_with_options():
    assert get_channel_options() == (("grpc.keepalive_time_ms", 45000),)


def test_overrides_default_values_if_provided():
    grpc_options = {"grpc.keepalive_time_ms": 4000}

    assert get_channel_options(grpc_options) == (("grpc.keepalive_time_ms", 4000),)


def test_adds_custom_options():
    grpc_options = {
        "grpc.keepalive_timeout_ms": 120000,
        "grpc.http2.min_time_between_pings_ms": 60000,
    }

    assert get_channel_options(grpc_options) == (
        ("grpc.keepalive_time_ms", 45000),
        ("grpc.keepalive_timeout_ms", 120000),
        ("grpc.http2.min_time_between_pings_ms", 60000),
    )



from typing import Dict

from pyzeebe.errors import NoZeebeAdapterError
from pyzeebe.job.job_status import JobStatus


class Job(object):
    def __init__(self, key: int, _type: str, process_instance_key: int, bpmn_process_id: str,
                 process_definition_version: int, process_definition_key: int, element_id: str, element_instance_key: int,
                 custom_headers: Dict, worker: str, retries: int, deadline: int, variables: Dict,
                 status: JobStatus = JobStatus.Running, zeebe_adapter=None):
        self.key = key
        self.type = _type
        self.process_instance_key = process_instance_key
        self.bpmn_process_id = bpmn_process_id
        self.process_definition_version = process_definition_version
        self.process_definition_key = process_definition_key
        self.element_id = element_id
        self.element_instance_key = element_instance_key
        self.custom_headers = custom_headers
        self.worker = worker
        self.retries = retries
        self.deadline = deadline
        self.variables = variables
        self.status = status
        self.zeebe_adapter = zeebe_adapter

    async def set_success_status(self) -> None:
        """
        Success status means that the job has been completed as intended.

        Raises:
            NoZeebeAdapterError: If the job does not have a configured ZeebeAdapter
            ZeebeBackPressureError: If Zeebe is currently in back pressure (too many requests)
            ZeebeGatewayUnavailableError: If the Zeebe gateway is unavailable
            ZeebeInternalError: If Zeebe experiences an internal error

        """
        if self.zeebe_adapter:
            await self.zeebe_adapter.complete_job(job_key=self.key, variables=self.variables)
        else:
            raise NoZeebeAdapterError()

    async def set_failure_status(self, message: str) -> None:
        """
        Failure status means a technical error has occurred. If retried the job may succeed.
        For example: connection to DB lost

        Args:
            message (str): The failure message that Zeebe will receive

        Raises:
            NoZeebeAdapterError: If the job does not have a configured ZeebeAdapter
            ZeebeBackPressureError: If Zeebe is currently in back pressure (too many requests)
            ZeebeGatewayUnavailableError: If the Zeebe gateway is unavailable
            ZeebeInternalError: If Zeebe experiences an internal error

        """
        if self.zeebe_adapter:
            await self.zeebe_adapter.fail_job(job_key=self.key, retries=self.retries-1, message=message)
        else:
            raise NoZeebeAdapterError()

    async def set_error_status(self, message: str, error_code: str = "") -> None:
        """
        Error status means that the job could not be completed because of a business error and won't ever be able to be completed.
        For example: a required parameter was not given
        An error code can be added to handle the error in the Zeebe process

        Args:
            message (str): The error message
            error_code (str): The error code that Zeebe will receive

        Raises:
            NoZeebeAdapterError: If the job does not have a configured ZeebeAdapter
            ZeebeBackPressureError: If Zeebe is currently in back pressure (too many requests)
            ZeebeGatewayUnavailableError: If the Zeebe gateway is unavailable
            ZeebeInternalError: If Zeebe experiences an internal error

        """
        if self.zeebe_adapter:
            await self.zeebe_adapter.throw_error(job_key=self.key, message=message, error_code=error_code)
        else:
            raise NoZeebeAdapterError()

    def __repr__(self):
        return str({"jobKey": self.key, "taskType": self.type, "processInstanceKey": self.process_instance_key,
                    "bpmnProcessId": self.bpmn_process_id,
                    "processDefinitionVersion": self.process_definition_version, "processDefinitionKey": self.process_definition_key,
                    "elementId": self.element_id, "elementInstanceKey": self.element_instance_key,
                    "customHeaders": self.custom_headers, "worker": self.worker, "retries": self.retries,
                    "deadline": self.deadline, "variables": self.variables})



import pytest

from pyzeebe.function_tools import dict_tools


@pytest.mark.asyncio
class TestConvertToDictFunction:
    async def test_converting_to_dict(self):
        async def original_function(x):
            return x

        dict_function = dict_tools.convert_to_dict_function(
            original_function, "x"
        )

        assert {"x": 1} == await dict_function(1)



from pyzeebe.errors.pyzeebe_errors import PyZeebeError


class ActivateJobsRequestInvalidError(PyZeebeError):
    def __init__(self, task_type: str, worker: str, timeout: int, max_jobs_to_activate: int):
        msg = "Failed to activate jobs. Reasons:"
        if task_type == "" or task_type is None:
            msg = msg + "task_type is empty, "
        if worker == "" or task_type is None:
            msg = msg + "worker is empty, "
        if timeout < 1:
            msg = msg + "job timeout is smaller than 0ms, "
        if max_jobs_to_activate < 1:
            msg = msg + "max_jobs_to_activate is smaller than 0ms, "

        super().__init__(msg)


class JobAlreadyDeactivatedError(PyZeebeError):
    def __init__(self, job_key: int):
        super().__init__(f"Job {job_key} was already stopped (Completed/Failed/Error)")
        self.job_key = job_key


class JobNotFoundError(PyZeebeError):
    def __init__(self, job_key: int):
        super().__init__(f"Job {job_key} not found")
        self.job_key = job_key



# type: ignore

import asyncio
from typing import Dict, List, Tuple

import grpc

from pyzeebe import ZeebeClient


class SyncZeebeClient(ZeebeClient):
    def __init__(
        self, grpc_channel: grpc.aio.Channel, max_connection_retries: int = 10
    ):
        super().__init__(grpc_channel, max_connection_retries)
        self.loop = asyncio.get_event_loop()

    def run_process(
        self, bpmn_process_id: str, variables: Dict = None, version: int = -1
    ) -> int:
        return self.loop.run_until_complete(
            super().run_process(bpmn_process_id, variables, version)
        )

    def run_process_with_result(
        self,
        bpmn_process_id: str,
        variables: Dict = None,
        version: int = -1,
        timeout: int = 0,
        variables_to_fetch: List[str] = None,
    ) -> Tuple[int, Dict]:
        return self.loop.run_until_complete(
            super().run_process_with_result(
                bpmn_process_id, variables, version, timeout, variables_to_fetch
            )
        )

    def cancel_process_instance(self, process_instance_key: int) -> int:
        return self.loop.run_until_complete(
            super().cancel_process_instance(process_instance_key)
        )

    def deploy_process(self, *process_file_path: str) -> None:
        return self.loop.run_until_complete(super().deploy_process(*process_file_path))

    def publish_message(
        self,
        name: str,
        correlation_key: str,
        variables: Dict = None,
        time_to_live_in_milliseconds: int = 60000,
        message_id: str = None,
    ) -> None:
        return self.loop.run_until_complete(
            super().publish_message(
                name,
                correlation_key,
                variables,
                time_to_live_in_milliseconds,
                message_id,
            )
        )



import logging

from pyzeebe import Job

logger = logging.getLogger(__name__)


class TaskState:
    def __init__(self):
        self._active_jobs = set()

    def remove(self, job: Job) -> None:
        try:
            self._active_jobs.remove(job.key)
        except KeyError:
            logger.warning(
                f"Could not find Job key {job.key} when trying to remove from TaskState")

    def add(self, job: Job) -> None:
        if job.key in self._active_jobs:
            raise ValueError(f"Job {job.key} already registered in TaskState")
        self._active_jobs.add(job.key)

    def count_active(self) -> int:
        return len(self._active_jobs)



from pyzeebe.grpc_internals.zeebe_job_adapter import ZeebeJobAdapter
from pyzeebe.grpc_internals.zeebe_message_adapter import ZeebeMessageAdapter
from pyzeebe.grpc_internals.zeebe_process_adapter import ZeebeProcessAdapter


# Mixin class
class ZeebeAdapter(ZeebeProcessAdapter, ZeebeJobAdapter, ZeebeMessageAdapter):
    pass



import pytest

from pyzeebe.worker.task_state import TaskState


@pytest.fixture
def task_state():
    return TaskState()


def test_new_task_state_has_0_active_jobs(task_state):
    assert task_state.count_active() == 0


def test_add_counts_as_active(task_state, job_from_task):
    task_state.add(job_from_task)
    assert task_state.count_active() == 1


def test_add_then_remove_results_in_0_active(task_state, job_from_task):
    task_state.add(job_from_task)
    task_state.remove(job_from_task)
    assert task_state.count_active() == 0


def test_prevents_adding_duplicate_job(task_state, job_from_task):
    with pytest.raises(ValueError) as exc_info:
        for _ in range(2):
            task_state.add(job_from_task)
    assert exc_info.value.args[0].endswith("already registered in TaskState")


def test_remove_non_existing_job_dont_withdraw_from_active_jobs(task_state, job_from_task):
    task_state.remove(job_from_task)
    assert task_state.count_active() == 0



import asyncio
import functools
from typing import List

from pyzeebe.function_tools import AsyncFunction, Function, SyncFunction


def asyncify_all_functions(functions: List[Function]) -> List[AsyncFunction]:
    async_functions = []
    for function in functions:
        if not is_async_function(function):
            async_functions.append(asyncify(function))
        else:
            # Mypy doesn't correctly understand that this is an async function
            async_functions.append(function)  # type: ignore
    return async_functions


def asyncify(task_function: SyncFunction) -> AsyncFunction:
    @functools.wraps(task_function)
    async def async_function(*args, **kwargs):
        loop = asyncio.get_event_loop()
        return await loop.run_in_executor(None, functools.partial(task_function, *args, **kwargs))
    return async_function


def is_async_function(function: Function) -> bool:
    # Not using inspect.iscoroutinefunction here because it doens't handle AsyncMock well
    # See: https://bugs.python.org/issue40573
    return asyncio.iscoroutinefunction(function)



from pyzeebe.errors.pyzeebe_errors import PyZeebeError


class InvalidOAuthCredentialsError(PyZeebeError):
    def __init__(self, url: str, client_id: str, audience: str):
        super().__init__(
            f"Invalid OAuth credentials supplied for {url} with audience {audience} and client id {client_id}")


class InvalidCamundaCloudCredentialsError(PyZeebeError):
    def __init__(self, client_id: str, cluster_id: str):
        super().__init__(f"Invalid credentials supplied for cluster {cluster_id} with client {client_id}")



from typing import Callable

import pytest

from pyzeebe import Job, TaskDecorator
from pyzeebe.task import task_builder
from pyzeebe.task.task import Task
from pyzeebe.task.task_config import TaskConfig


class TestBuildTask:
    @pytest.fixture
    def single_value_task_config(self, task_config: TaskConfig):
        task_config.single_value = True
        task_config.variable_name = "y"

        return task_config

    def test_returns_task(self, original_task_function: Callable, task_config: TaskConfig):
        task = task_builder.build_task(original_task_function, task_config)

        assert isinstance(task, Task)

    @pytest.mark.asyncio
    async def test_single_value_func(self, single_value_task_config: TaskConfig, mocked_job_with_adapter: Job):
        task = task_builder.build_task(lambda: 1, single_value_task_config)
        job = await task.job_handler(mocked_job_with_adapter)

        assert job.variables.pop("y") == 1

    @pytest.mark.asyncio
    async def test_no_additional_variables_are_added_to_result(self, single_value_task_config: TaskConfig, mocked_job_with_adapter: Job):
        mocked_job_with_adapter.variables = {"x": 1}

        task = task_builder.build_task(lambda x: x, single_value_task_config)
        job = await task.job_handler(mocked_job_with_adapter)

        assert len(job.variables.keys()) == 1
        assert set(job.variables.keys()) == {"y"}

    @pytest.mark.asyncio
    async def test_job_parameter_is_injected_in_task(self, task_config: TaskConfig, mocked_job_with_adapter: Job):
        def function_with_job_parameter(job: Job):
            return {"job": job}

        task = task_builder.build_task(
            function_with_job_parameter, task_config)
        job = await task.job_handler(mocked_job_with_adapter)

        assert job.variables["job"] == mocked_job_with_adapter


class TestBuildJobHandler:
    def test_returned_task_is_callable(self, original_task_function: Callable, task_config: TaskConfig):
        task = task_builder.build_job_handler(
            original_task_function, task_config)
        assert callable(task)

    @pytest.mark.asyncio
    async def test_exception_handler_called(self, original_task_function: Callable, task_config: TaskConfig,
                                            mocked_job_with_adapter: Job):
        exception = Exception()
        original_task_function.side_effect = exception
        job_handler = task_builder.build_job_handler(
            original_task_function, task_config)

        await job_handler(mocked_job_with_adapter)

        task_config.exception_handler.assert_called_with(
            exception, mocked_job_with_adapter)

    @pytest.mark.asyncio
    async def test_parameters_are_provided_to_task(self, original_task_function: Callable, task_config: TaskConfig,
                                                   mocked_job_with_adapter: Job):
        mocked_job_with_adapter.variables = {"x": 1}
        job_handler = task_builder.build_job_handler(
            original_task_function, task_config)

        await job_handler(mocked_job_with_adapter)

        original_task_function.assert_called_with(x=1)

    @pytest.mark.asyncio
    async def test_variables_are_added_to_result(self, original_task_function: Callable, task_config: TaskConfig,
                                                 mocked_job_with_adapter: Job):
        original_task_function.return_value = {"x": 1}
        job_handler = task_builder.build_job_handler(
            original_task_function, task_config
        )

        job = await job_handler(mocked_job_with_adapter)

        assert job.variables.pop("x") == 1

    @pytest.mark.asyncio
    async def test_complete_job_called(self, original_task_function: Callable, task_config: TaskConfig,
                                       mocked_job_with_adapter: Job):
        job_handler = task_builder.build_job_handler(
            original_task_function, task_config)

        await job_handler(mocked_job_with_adapter)

        mocked_job_with_adapter.set_success_status.assert_called_once()

    @pytest.mark.asyncio
    async def test_returned_task_runs_original_function(self, original_task_function: Callable, task_config: TaskConfig,
                                                        mocked_job_with_adapter: Job):
        job_handler = task_builder.build_job_handler(
            original_task_function, task_config)

        await job_handler(mocked_job_with_adapter)

        original_task_function.assert_called_once()

    @pytest.mark.asyncio
    async def test_before_decorator_called(self, original_task_function: Callable, decorator: TaskDecorator,
                                           task_config: TaskConfig,
                                           mocked_job_with_adapter: Job):
        task_config.before.append(decorator)
        job_handler = task_builder.build_job_handler(
            original_task_function, task_config
        )

        await job_handler(mocked_job_with_adapter)

        task_config.before.pop().assert_called_once()

    @pytest.mark.asyncio
    async def test_after_decorator_called(self, original_task_function: Callable, decorator: TaskDecorator,
                                          task_config: TaskConfig,
                                          mocked_job_with_adapter: Job):
        task_config.after.append(decorator)
        job_handler = task_builder.build_job_handler(
            original_task_function, task_config
        )

        await job_handler(mocked_job_with_adapter)

        task_config.after.pop().assert_called_once()

    @pytest.mark.asyncio
    async def test_failing_decorator_continues(self, original_task_function: Callable, decorator: TaskDecorator,
                                               task_config: TaskConfig, mocked_job_with_adapter: Job):
        decorator.side_effect = Exception()
        task_config.before.append(decorator)
        job_handler = task_builder.build_job_handler(
            original_task_function, task_config
        )

        await job_handler(mocked_job_with_adapter)

        decorator.assert_called_once()
        task_config.exception_handler.assert_not_called()

    @pytest.mark.asyncio
    async def test_decorator_variables_are_added(self, original_task_function: Callable, decorator: TaskDecorator,
                                                 task_config: TaskConfig, mocked_job_with_adapter: Job):
        mocked_job_with_adapter.variables = {"x": 2}
        decorator_return_value = mocked_job_with_adapter
        decorator.return_value = decorator_return_value
        job_handler = task_builder.build_job_handler(
            original_task_function, task_config)

        job = await job_handler(mocked_job_with_adapter)

        assert "x" in job.variables

    @pytest.mark.asyncio
    async def test_job_parameter_is_injected(self, task_config: TaskConfig, mocked_job_with_adapter: Job):
        def function_with_job_parameter(x: int, job: Job):
            return {"job": job}

        task_config.job_parameter_name = "job"

        job_handler = task_builder.build_job_handler(
            function_with_job_parameter, task_config
        )
        job = await job_handler(mocked_job_with_adapter)

        assert job.variables["job"] == mocked_job_with_adapter



__version__ = "3.0.0rc3"

from pyzeebe import errors
from pyzeebe.channel import *
from pyzeebe.client.client import ZeebeClient
from pyzeebe.client.sync_client import SyncZeebeClient  # type: ignore
from pyzeebe.job.job import Job
from pyzeebe.job.job_status import JobStatus
from pyzeebe.task.exception_handler import ExceptionHandler
from pyzeebe.task.task_config import TaskConfig
from pyzeebe.task.types import TaskDecorator
from pyzeebe.worker.task_router import (ZeebeTaskRouter,
                                        default_exception_handler)
from pyzeebe.worker.worker import ZeebeWorker



""" gRPC Channel Options

``grpc.keepalive_time_ms``
--------------------------

Time between keepalive pings. Following the official Zeebe Java/Go client, sending pings every 45 seconds.

https://docs.camunda.io/docs/product-manuals/zeebe/deployment-guide/operations/setting-up-a-cluster/#keep-alive-intervals
"""
from typing import Any, Dict, Tuple

GRPC_CHANNEL_OPTIONS = {
    "grpc.keepalive_time_ms": 45_000
}


def get_channel_options(options: Dict[str, Any] = None) -> Tuple[Tuple[str, Any], ...]:
    """
    Convert options dict to tuple in expected format for creating the gRPC channel

    Args:
        options (Dict[str, Any]): A key/value representation of `gRPC channel arguments_`.
            Default: None (will use library defaults)

    Returns:
        Tuple[Tuple[str, Any], ...]: Options for the gRPC channel

    .. _gRPC channel arguments:
        https://grpc.github.io/grpc/python/glossary.html#term-channel_arguments
    """
    if options:
        options = {**GRPC_CHANNEL_OPTIONS, **options}
    else:
        options = GRPC_CHANNEL_OPTIONS
    return tuple(
        (k, v) for k, v in options.items()
    )



from typing import Callable, List

from pyzeebe.function_tools import async_tools
from pyzeebe.job.job import Job
from pyzeebe.task.task_config import TaskConfig
from tests.unit.utils.function_tools import functions_are_all_async


class TestConstructor:
    async def async_decorator(self, job: Job) -> Job:
        return job

    def sync_decorator(self, job: Job) -> Job:
        return job

    def exception_handler(self, ):
        pass

    def test_before_decorators_are_async(self, task_type: str):
        task_config = TaskConfig(
            task_type,
            self.exception_handler,
            10000,
            32,
            [],
            False,
            "",
            [self.sync_decorator, self.async_decorator],
            []
        )

        assert functions_are_all_async(task_config.before)

    def test_after_decorators_are_async(self, task_type: str):
        task_config = TaskConfig(
            task_type,
            self.exception_handler,
            10000,
            32,
            [],
            False,
            "",
            [],
            [self.sync_decorator, self.async_decorator]
        )

        assert functions_are_all_async(task_config.after)



from pyzeebe.channel.camunda_cloud_channel import create_camunda_cloud_channel
from pyzeebe.channel.insecure_channel import create_insecure_channel
from pyzeebe.channel.secure_channel import create_secure_channel



from random import randint
from uuid import uuid4

import pytest
from zeebe_grpc.gateway_pb2 import (CompleteJobResponse, FailJobResponse,
                                    ThrowErrorResponse)

from pyzeebe.errors import (ActivateJobsRequestInvalidError,
                            JobAlreadyDeactivatedError, JobNotFoundError)
from pyzeebe.grpc_internals.zeebe_job_adapter import ZeebeJobAdapter
from pyzeebe.job.job import Job
from pyzeebe.task.task import Task
from tests.unit.utils.gateway_mock import GatewayMock
from tests.unit.utils.random_utils import RANDOM_RANGE, random_job


def random_job_key() -> int:
    return randint(0, RANDOM_RANGE)


def random_message() -> str:
    return str(uuid4())


@pytest.mark.asyncio
class TestActivateJobs:
    zeebe_job_adapter: ZeebeJobAdapter

    @pytest.fixture(autouse=True)
    def set_up(self, zeebe_adapter: ZeebeJobAdapter):
        self.zeebe_job_adapter = zeebe_adapter

    def activate_jobs(
        self,
        task_type=str(uuid4()),
        worker=str(uuid4()),
        timeout=randint(10, 100),
        request_timeout=100,
        max_jobs_to_activate=1,
        variables_to_fetch=[]
    ):
        return self.zeebe_job_adapter.activate_jobs(task_type, worker, timeout, max_jobs_to_activate, variables_to_fetch, request_timeout)

    async def test_returns_correct_amount_of_jobs(self, grpc_servicer: GatewayMock, task: Task):
        active_jobs_count = randint(4, 100)
        for _ in range(0, active_jobs_count):
            job = random_job(task)
            grpc_servicer.active_jobs[job.key] = job

        jobs = self.activate_jobs(task_type=task.type)

        assert len([job async for job in jobs]) == active_jobs_count

    async def test_raises_on_invalid_worker(self):
        with pytest.raises(ActivateJobsRequestInvalidError):
            jobs = self.activate_jobs(worker=None)
            await jobs.__anext__()

    async def test_raises_on_invalid_job_timeout(self):
        with pytest.raises(ActivateJobsRequestInvalidError):
            jobs = self.activate_jobs(timeout=0)
            await jobs.__anext__()

    async def test_raises_on_invalid_task_type(self):
        with pytest.raises(ActivateJobsRequestInvalidError):
            jobs = self.activate_jobs(task_type=None)
            await jobs.__anext__()

    async def test_raises_on_invalid_max_jobs(self):
        with pytest.raises(ActivateJobsRequestInvalidError):
            jobs = self.activate_jobs(max_jobs_to_activate=0)
            await jobs.__anext__()


@pytest.mark.asyncio
class TestCompleteJob:
    async def test_response_is_of_correct_type(self, zeebe_adapter: ZeebeJobAdapter, first_active_job: Job):
        response = await zeebe_adapter.complete_job(first_active_job.key, {})

        assert isinstance(response, CompleteJobResponse)

    async def test_raises_on_fake_job(self, zeebe_adapter: ZeebeJobAdapter):
        with pytest.raises(JobNotFoundError):
            await zeebe_adapter.complete_job(random_job_key(), {})

    async def test_raises_on_already_completed_job(self, zeebe_adapter: ZeebeJobAdapter, first_active_job: Job):
        await zeebe_adapter.complete_job(first_active_job.key, {})

        with pytest.raises(JobAlreadyDeactivatedError):
            await zeebe_adapter.complete_job(first_active_job.key, {})


@pytest.mark.asyncio
class TestFailJob:
    async def test_response_is_of_correct_type(self, zeebe_adapter: ZeebeJobAdapter, first_active_job: Job):
        response = await zeebe_adapter.fail_job(first_active_job.key, first_active_job.retries, random_message())

        assert isinstance(response, FailJobResponse)

    async def test_raises_on_fake_job(self, zeebe_adapter: ZeebeJobAdapter):
        with pytest.raises(JobNotFoundError):
            await zeebe_adapter.fail_job(random_job_key(), 1, random_message())

    async def test_raises_on_deactivated_job(self, zeebe_adapter: ZeebeJobAdapter, first_active_job: Job):
        await zeebe_adapter.fail_job(first_active_job.key, first_active_job.retries, random_message())

        with pytest.raises(JobAlreadyDeactivatedError):
            await zeebe_adapter.fail_job(first_active_job.key, first_active_job.retries, random_message())


@pytest.mark.asyncio
class TestThrowError:
    async def test_response_is_of_correct_type(self, zeebe_adapter: ZeebeJobAdapter, first_active_job: Job):
        response = await zeebe_adapter.throw_error(first_active_job.key, random_message())

        assert isinstance(response, ThrowErrorResponse)

    async def test_raises_on_fake_job(self, zeebe_adapter: ZeebeJobAdapter):
        with pytest.raises(JobNotFoundError):
            await zeebe_adapter.throw_error(random_job_key(), random_message())

    async def test_raises_on_deactivated_job(self, zeebe_adapter: ZeebeJobAdapter, first_active_job: Job):
        await zeebe_adapter.throw_error(first_active_job.key, random_message())

        with pytest.raises(JobAlreadyDeactivatedError):
            await zeebe_adapter.throw_error(first_active_job.key, random_message())



import functools
import logging
from typing import Awaitable, Callable, Dict, Sequence, Tuple

from pyzeebe import Job
from pyzeebe.function_tools import AsyncFunction, DictFunction, Function
from pyzeebe.function_tools.async_tools import asyncify, is_async_function
from pyzeebe.function_tools.dict_tools import convert_to_dict_function
from pyzeebe.function_tools.parameter_tools import get_job_parameter_name
from pyzeebe.task.task import Task
from pyzeebe.task.task_config import TaskConfig
from pyzeebe.task.types import AsyncTaskDecorator, DecoratorRunner, JobHandler

logger = logging.getLogger(__name__)


def build_task(task_function: Function, task_config: TaskConfig) -> Task:
    task_config.job_parameter_name = get_job_parameter_name(task_function)
    return Task(task_function, build_job_handler(task_function, task_config), task_config)


def build_job_handler(task_function: Function, task_config: TaskConfig) -> JobHandler:
    prepared_task_function = prepare_task_function(task_function, task_config)

    before_decorator_runner = create_decorator_runner(task_config.before)
    after_decorator_runner = create_decorator_runner(task_config.after)

    @functools.wraps(task_function)
    async def job_handler(job: Job) -> Job:
        if task_config.job_parameter_name:
            job.variables[task_config.job_parameter_name] = job

        job = await before_decorator_runner(job)
        job.variables, succeeded = await run_original_task_function(
            prepared_task_function, task_config, job
        )
        job = await after_decorator_runner(job)
        if succeeded:
            await job.set_success_status()
        return job

    return job_handler


def prepare_task_function(task_function: Function, task_config: TaskConfig) -> DictFunction:
    if not is_async_function(task_function):
        task_function = asyncify(task_function)

    if task_config.single_value:
        task_function = convert_to_dict_function(
            task_function, task_config.variable_name
        )
    return task_function


async def run_original_task_function(task_function: DictFunction, task_config: TaskConfig, job: Job) -> Tuple[Dict, bool]:
    try:
        return await task_function(**job.variables), True  # type: ignore
    except Exception as e:
        logger.debug(f"Failed job: {job}. Error: {e}.")
        await task_config.exception_handler(e, job)
        return job.variables, False


def create_decorator_runner(decorators: Sequence[AsyncTaskDecorator]) -> DecoratorRunner:
    async def decorator_runner(job: Job):
        for decorator in decorators:
            job = await run_decorator(decorator, job)
        return job

    return decorator_runner


async def run_decorator(decorator: AsyncTaskDecorator, job: Job) -> Job:
    try:
        return await decorator(job)
    except Exception as e:
        logger.warning(f"Failed to run decorator {decorator}. Exception: {e}")
        return job



from typing import Callable, List

import pytest

from pyzeebe.function_tools import parameter_tools
from pyzeebe.job.job import Job
from tests.unit.utils import dummy_functions


class TestGetFunctionParameters:
    @pytest.mark.parametrize("fn,expected", [
        (dummy_functions.no_param, []),
        (dummy_functions.one_param, ["x"]),
        (dummy_functions.multiple_params, ["x", "y", "z"]),
        (dummy_functions.one_keyword_param, ["x"]),
        (dummy_functions.multiple_keyword_param, ["x", "y", "z"]),
        (dummy_functions.positional_and_keyword_params, ["x", "y"]),
        (dummy_functions.args_param, []),
        (dummy_functions.kwargs_param, []),
        (dummy_functions.standard_named_params, ["args", "kwargs"]),
        (dummy_functions.lambda_no_params, []),
        (dummy_functions.lambda_one_param, ["x"]),
        (dummy_functions.lambda_multiple_params, ["x", "y", "z"]),
        (dummy_functions.lambda_one_keyword_param, ["x"]),
        (dummy_functions.lambda_multiple_keyword_params, ["x", "y", "z"]),
        (dummy_functions.lambda_positional_and_keyword_params, ["x", "y"])
    ])
    def test_get_params(self, fn: Callable, expected: List[str]):
        assert parameter_tools.get_parameters_from_function(fn) == expected


class TestGetJobParameter:
    def test_returns_none_when_there_are_no_parameters_annotated_with_job(self):
        job_parameter = parameter_tools.get_job_parameter_name(
            dummy_functions.multiple_params)

        assert job_parameter == None

    def test_returns_parameter_name_when_annotated(self):
        job_parameter = parameter_tools.get_job_parameter_name(
            dummy_functions.with_job_parameter)

        assert job_parameter == "job"

    def test_returns_first_parameter_annotated_with_job(self):
        job_parameter = parameter_tools.get_job_parameter_name(
            dummy_functions.with_multiple_job_parameters)

        assert job_parameter == "job"



import grpc
import pytest
from mock import AsyncMock

from pyzeebe.errors import (ZeebeBackPressureError,
                            ZeebeGatewayUnavailableError, ZeebeInternalError)
from pyzeebe.grpc_internals.zeebe_adapter_base import ZeebeAdapterBase


def test_should_retry_no_current_retries(zeebe_adapter: ZeebeAdapterBase):
    zeebe_adapter._max_connection_retries = 1
    assert zeebe_adapter._should_retry()


def test_should_retry_current_retries_over_max(zeebe_adapter: ZeebeAdapterBase):
    zeebe_adapter._max_connection_retries = 1
    zeebe_adapter._current_connection_retries = 1
    assert not zeebe_adapter._should_retry()


@pytest.mark.asyncio
async def test_common_zeebe_grpc_error_internal(zeebe_adapter: ZeebeAdapterBase):
    error = grpc.aio.AioRpcError(
        grpc.StatusCode.INTERNAL, None, None
    )
    with pytest.raises(ZeebeInternalError):
        await zeebe_adapter._common_zeebe_grpc_errors(error)


@pytest.mark.asyncio
async def test_common_zeebe_grpc_error_back_pressure(zeebe_adapter: ZeebeAdapterBase):
    error = grpc.aio.AioRpcError(
        grpc.StatusCode.RESOURCE_EXHAUSTED, None, None
    )
    with pytest.raises(ZeebeBackPressureError):
        await zeebe_adapter._common_zeebe_grpc_errors(error)


@pytest.mark.asyncio
async def test_common_zeebe_grpc_error_gateway_unavailable(zeebe_adapter: ZeebeAdapterBase):
    error = grpc.aio.AioRpcError(
        grpc.StatusCode.UNAVAILABLE, None, None
    )
    with pytest.raises(ZeebeGatewayUnavailableError):
        await zeebe_adapter._common_zeebe_grpc_errors(error)


@pytest.mark.asyncio
async def test_common_zeebe_grpc_error_unkown_error(zeebe_adapter: ZeebeAdapterBase):
    error = grpc.aio.AioRpcError(
        "FakeGrpcStatus", None, None
    )
    with pytest.raises(grpc.aio.AioRpcError):
        await zeebe_adapter._common_zeebe_grpc_errors(error)


@pytest.mark.asyncio
async def test_close_after_retried_unavailable(zeebe_adapter: ZeebeAdapterBase):
    error = grpc.aio.AioRpcError(
        grpc.StatusCode.UNAVAILABLE, None, None
    )

    zeebe_adapter._close = AsyncMock()
    zeebe_adapter._max_connection_retries = 1
    with pytest.raises(ZeebeGatewayUnavailableError):
        await zeebe_adapter._common_zeebe_grpc_errors(error)

    zeebe_adapter._close.assert_called_once()


@pytest.mark.asyncio
async def test_close_after_retried_internal(zeebe_adapter: ZeebeAdapterBase):
    error = grpc.aio.AioRpcError(
        grpc.StatusCode.INTERNAL, None, None
    )
    zeebe_adapter._close = AsyncMock()
    zeebe_adapter._max_connection_retries = 1
    with pytest.raises(ZeebeInternalError):
        await zeebe_adapter._common_zeebe_grpc_errors(error)

    zeebe_adapter._close.assert_called_once()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pyzeebe",
    version="3.0.0rc3",
    author="Jonatan Martens",
    author_email="jonatanmartenstav@gmail.com",
    description="Zeebe client api",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/JonatanMartens/pyzeebe",
    packages=setuptools.find_packages(exclude=("tests",)),
    install_requires=["oauthlib==3.1.0",
                      "requests-oauthlib==1.3.0", "zeebe-grpc==1.0.0",
                      "aiofiles==0.7.0"],
    exclude=["*test.py", "tests", "*.bpmn"],
    keywords="zeebe workflow workflow-engine",
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)



from random import randint
from uuid import uuid4

import pytest


@pytest.fixture
def deployed_process(grpc_servicer):
    bpmn_process_id = str(uuid4())
    version = randint(0, 10)
    grpc_servicer.mock_deploy_process(bpmn_process_id, version, [])
    return bpmn_process_id, version



import os
from uuid import uuid4

from pyzeebe.channel.utils import DEFAULT_ADDRESS, create_address


class TestCreateAddress:
    def test_returns_default_address(self):
        address = create_address()

        assert address == DEFAULT_ADDRESS

    def test_default_port_is_26500(self):
        address = create_address(hostname=str(uuid4()))

        assert address.split(":")[1] == "26500"

    def test_default_hostname_is_localhost(self):
        address = create_address(port=12)

        assert address.split(":")[0] == "localhost"

    def test_returns_env_var_if_provided(self):
        zeebe_address = str(uuid4())
        os.environ["ZEEBE_ADDRESS"] = zeebe_address

        address = create_address()

        assert address == zeebe_address



from random import randint
from uuid import uuid4

import pytest
from zeebe_grpc.gateway_pb2 import PublishMessageResponse

from pyzeebe.errors import MessageAlreadyExistsError
from pyzeebe.grpc_internals.zeebe_message_adapter import ZeebeMessageAdapter
from tests.unit.utils.random_utils import RANDOM_RANGE


@pytest.mark.asyncio
class TestPublishMessage:
    zeebe_message_adapter: ZeebeMessageAdapter

    @pytest.fixture(autouse=True)
    def set_up(self, zeebe_adapter: ZeebeMessageAdapter):
        self.zeebe_message_adapter = zeebe_adapter

    async def publish_message(
        self,
        name=str(uuid4()),
        variables={},
        correlation_key=str(uuid4()),
        time_to_live_in_milliseconds=randint(0, RANDOM_RANGE),
        message_id=str(uuid4())
    ):
        return await self.zeebe_message_adapter.publish_message(name, correlation_key, time_to_live_in_milliseconds, variables, message_id)

    async def test_response_is_of_correct_type(self):
        response = await self.publish_message()

        assert isinstance(response, PublishMessageResponse)

    async def test_raises_on_invalid_name(self):
        with pytest.raises(TypeError):
            await self.publish_message(name=randint(0, RANDOM_RANGE))

    async def test_raises_on_invalid_correlation_key(self):
        with pytest.raises(TypeError):
            await self.publish_message(correlation_key=randint(0, RANDOM_RANGE))

    async def test_raises_on_invalid_time_to_live(self):
        with pytest.raises(TypeError):
            await self.publish_message(time_to_live_in_milliseconds=str(uuid4()))

    async def test_raises_on_duplicate_message(self):
        message_id = str(uuid4())

        with pytest.raises(MessageAlreadyExistsError):
            await self.publish_message(message_id=message_id)
            await self.publish_message(message_id=message_id)



from typing import Callable

from pyzeebe.task.task_config import TaskConfig
from pyzeebe.task.types import JobHandler


class Task:
    def __init__(self, original_function: Callable, job_handler: JobHandler, config: TaskConfig):
        self.original_function = original_function
        self.job_handler = job_handler
        self.config = config

    @property
    def type(self):
        return self.config.type

    def __repr__(self):
        return f"Task(config= {self.config}, original_function={self.original_function}, " \
               f"job_handler={self.job_handler})"



from typing import Dict, Optional

import grpc

from pyzeebe.channel.channel_options import get_channel_options
from pyzeebe.channel.utils import create_address


def create_secure_channel(
    hostname: Optional[str] = None,
    port: Optional[int] = None,
    channel_options: Optional[Dict] = None,
    channel_credentials: Optional[grpc.ChannelCredentials] = None,
) -> grpc.aio.Channel:
    """
    Create a secure channel

    Args:
        hostname (Optional[str], optional): Zeebe gateway hostname
        port (Optional[int], optional): Zeebe gateway port
        channel_options (Optional[Dict], optional): GRPC channel options. See https://grpc.github.io/grpc/python/glossary.html#term-channel_arguments
        channel_credentials (Optional[grpc.ChannelCredentials]): Channel credentials to use. Will use grpc.ssl_channel_credentials() if not provided.

    Returns:
        grpc.aio.Channel: A GRPC Channel connected to the Zeebe gateway.
    """
    address = create_address(hostname, port)
    credentials = channel_credentials or grpc.ssl_channel_credentials()
    return grpc.aio.secure_channel(
        address, credentials, options=get_channel_options(channel_options)
    )



import asyncio
import logging
import socket
from typing import List, Optional

import grpc

from pyzeebe import TaskDecorator
from pyzeebe.grpc_internals.zeebe_adapter import ZeebeAdapter
from pyzeebe.worker.job_executor import JobExecutor
from pyzeebe.worker.job_poller import JobPoller
from pyzeebe.worker.task_router import ZeebeTaskRouter
from pyzeebe.worker.task_state import TaskState

logger = logging.getLogger(__name__)


class ZeebeWorker(ZeebeTaskRouter):
    """A zeebe worker that can connect to a zeebe instance and perform tasks."""

    def __init__(
        self,
        grpc_channel: grpc.aio.Channel,
        name: Optional[str] = None,
        request_timeout: int = 0,
        before: List[TaskDecorator] = None,
        after: List[TaskDecorator] = None,
        max_connection_retries: int = 10,
        watcher_max_errors_factor: int = 3,
        max_task_count: int = 32,
    ):
        """
        Args:
            grpc_channel (grpc.aio.Channel): GRPC Channel connected to a Zeebe gateway
            name (str): Name of zeebe worker
            request_timeout (int): Longpolling timeout for getting tasks from zeebe. If 0 default value is used
            before (List[TaskDecorator]): Decorators to be performed before each task
            after (List[TaskDecorator]): Decorators to be performed after each task
            max_connection_retries (int): Amount of connection retries before worker gives up on connecting to zeebe. To setup with infinite retries use -1
            watcher_max_errors_factor (int): Number of consequtive errors for a task watcher will accept before raising MaxConsecutiveTaskThreadError
            max_task_count (int): The maximum amount of tasks the worker can handle simultaniously
        """
        super().__init__(before, after)
        self.zeebe_adapter = ZeebeAdapter(grpc_channel, max_connection_retries)
        self.name = name or socket.gethostname()
        self.request_timeout = request_timeout
        self.watcher_max_errors_factor = watcher_max_errors_factor
        self._watcher_thread = None
        self.max_task_count = max_task_count
        self._task_state = TaskState()
        self._job_pollers: List[JobPoller] = []
        self._job_executors: List[JobExecutor] = []

    async def work(self) -> None:
        """
        Start the worker. The worker will poll zeebe for jobs of each task in a different thread.

        Raises:
            ActivateJobsRequestInvalidError: If one of the worker's task has invalid types
            ZeebeBackPressureError: If Zeebe is currently in back pressure (too many requests)
            ZeebeGatewayUnavailableError: If the Zeebe gateway is unavailable
            ZeebeInternalError: If Zeebe experiences an internal error

        """
        self._job_executors, self._job_pollers = [], []

        for task in self.tasks:
            jobs_queue: asyncio.Queue = asyncio.Queue()
            poller = JobPoller(
                self.zeebe_adapter,
                task,
                jobs_queue,
                self.name,
                self.request_timeout,
                self._task_state,
                self.max_task_count,
            )
            executor = JobExecutor(task, jobs_queue, self._task_state)
            self._job_pollers.append(poller)
            self._job_executors.append(executor)

        coroutines = [poller.poll() for poller in self._job_pollers] + [
            executor.execute() for executor in self._job_executors
        ]

        return await asyncio.gather(*coroutines)

    async def stop(self) -> None:
        """
        Stop the worker. This will emit a signal asking tasks to complete the current task and stop polling for new.
        """
        for poller in self._job_pollers:
            await poller.stop()

        for executor in self._job_executors:
            await executor.stop()

    def include_router(self, *routers: ZeebeTaskRouter) -> None:
        """
        Adds all router's tasks to the worker.

        Raises:
            DuplicateTaskTypeError: If a task from the router already exists in the worker

        """
        for router in routers:
            for task in router.tasks:
                task.config = self._add_decorators_to_config(task.config)
                self._add_task(task)



from typing import Callable, List

from pyzeebe.function_tools import async_tools


def functions_are_all_async(functions: List[Callable]) -> bool:
    return all(
        [
            async_tools.is_async_function(function)
            for function
            in functions
        ]
    )



from pyzeebe.errors.pyzeebe_errors import PyZeebeError


class ProcessDefinitionNotFoundError(PyZeebeError):
    def __init__(self, bpmn_process_id: str, version: int):
        super().__init__(
            f"Process definition: {bpmn_process_id}  with {version} was not found")
        self.bpmn_process_id = bpmn_process_id
        self.version = version


class ProcessInstanceNotFoundError(PyZeebeError):
    def __init__(self, process_instance_key: int):
        super().__init__(
            f"Process instance key: {process_instance_key} was not found")
        self.process_instance_key = process_instance_key


class ProcessDefinitionHasNoStartEventError(PyZeebeError):
    def __init__(self, bpmn_process_id: str):
        super().__init__(
            f"Process {bpmn_process_id} has no start event that can be called manually")
        self.bpmn_process_id = bpmn_process_id


class ProcessInvalidError(PyZeebeError):
    pass


class InvalidJSONError(PyZeebeError):
    pass



from uuid import uuid4

import pytest
from mock import AsyncMock

from pyzeebe.errors import NoZeebeAdapterError


@pytest.mark.asyncio
async def test_success(job_with_adapter):
    complete_job_mock = AsyncMock()
    job_with_adapter.zeebe_adapter.complete_job = complete_job_mock

    await job_with_adapter.set_success_status()

    complete_job_mock.assert_called_with(
        job_key=job_with_adapter.key, variables=job_with_adapter.variables
    )


@pytest.mark.asyncio
async def test_success_no_zeebe_adapter(job_without_adapter):
    with pytest.raises(NoZeebeAdapterError):
        await job_without_adapter.set_success_status()


@pytest.mark.asyncio
async def test_error(job_with_adapter):
    throw_error_mock = AsyncMock()
    job_with_adapter.zeebe_adapter.throw_error = throw_error_mock
    message = str(uuid4())

    await job_with_adapter.set_error_status(message)

    throw_error_mock.assert_called_with(
        job_key=job_with_adapter.key, message=message, error_code=""
    )


@pytest.mark.asyncio
async def test_error_with_code(job_with_adapter):
    throw_error_mock = AsyncMock()
    job_with_adapter.zeebe_adapter.throw_error = throw_error_mock
    message = str(uuid4())
    error_code = "custom-error-code"

    await job_with_adapter.set_error_status(message, error_code)

    throw_error_mock.assert_called_with(
        job_key=job_with_adapter.key, message=message, error_code=error_code
    )


@pytest.mark.asyncio
async def test_error_no_zeebe_adapter(job_without_adapter):
    with pytest.raises(NoZeebeAdapterError):
        message = str(uuid4())
        await job_without_adapter.set_error_status(message)


@pytest.mark.asyncio
async def test_failure(job_with_adapter):
    fail_job_mock = AsyncMock()
    job_with_adapter.zeebe_adapter.fail_job = fail_job_mock
    message = str(uuid4())

    await job_with_adapter.set_failure_status(message)

    fail_job_mock.assert_called_with(
        job_key=job_with_adapter.key, retries=job_with_adapter.retries-1, message=message
    )


@pytest.mark.asyncio
async def test_failure_no_zeebe_adapter(job_without_adapter):
    with pytest.raises(NoZeebeAdapterError):
        message = str(uuid4())
        await job_without_adapter.set_failure_status(message)



from typing import Awaitable, Callable, Dict, TypeVar, Union

Parameters = TypeVar("Parameters")
ReturnType = TypeVar("ReturnType")

SyncFunction = Callable[[Parameters], ReturnType]
AsyncFunction = Callable[[Parameters], Awaitable[ReturnType]]
Function = Union[SyncFunction, AsyncFunction]

DictFunction = Callable[[Parameters], Awaitable[Dict]]
